"""Tests for the LALInference interface."""

import unittest
from unittest.mock import Mock, patch

import shutil
import os
import git

import io
import contextlib


from click.testing import CliRunner

from asimov.utils import set_directory
from asimov import config

from asimov.cli import project, manage
from asimov.cli import configuration
from asimov.cli.application import apply_page
from asimov.ledger import YAMLLedger
from asimov.event import Event
from asimov.pipeline import PipelineException
from asimov.pipelines.bayeswave import BayesWave
from asimov.event import Event
from asimov.pipeline import PipelineException

TEST_YAML = """
name: S000000xx
repository: {0}/tests/test_data/s000000xx/
working_directory: {0}/tests/tmp/s000000xx/
webdir: ''
productions:
- Prod1:
    rundir: {0}/tests/tmp/s000000xx/C01_offline/Prod1
    pipeline: bayeswave
    comment: PSD production
    status: wait

"""

class BayeswaveTests(unittest.TestCase):
    """Test bayeswave interface.

    TODO
    ----
    Right now these feel a bit more like an expression of intention than actual tests, as we'll need to set the testing environment up better to make this work.

    The test_dag method will need to be updated.
"""

    @classmethod
    def setUpClass(cls):
        cls.cwd = os.getcwd()

    def setUp(self):
        os.makedirs(f"{self.cwd}/tests/tmp/project")
        os.chdir(f"{self.cwd}/tests/tmp/project")
        runner = CliRunner()
        result = runner.invoke(project.init,
                               ['Test Project', '--root', f"{self.cwd}/tests/tmp/project"])
        assert result.exit_code == 0
        assert result.output == '● New project created successfully!\n'
        self.ledger = YAMLLedger(f".asimov/ledger.yml")

    def tearDown(self):
        os.chdir(self.cwd)
        shutil.rmtree(f"{self.cwd}/tests/tmp/project/")

    @unittest.skip("Skipped temporarily while RIFT is updated")
    def test_build_cli(self):
        """Check that a RIFT config file can be built."""
        apply_page(file = "https://git.ligo.org/asimov/data/-/raw/main/defaults/production-pe.yaml", event=None, ledger=self.ledger)
        apply_page(file = "https://git.ligo.org/asimov/data/-/raw/main/defaults/production-pe-priors.yaml", event=None, ledger=self.ledger)
        event = "GW150914_095045"
        pipeline = "bayeswave"
        apply_page(file = f"https://git.ligo.org/asimov/data/-/raw/main/tests/{event}.yaml", event=None, ledger=self.ledger)
        apply_page(file = f"https://git.ligo.org/asimov/data/-/raw/main/tests/{pipeline}.yaml", event=event, ledger=self.ledger)

        runner = CliRunner()
        result = runner.invoke(manage.build, "--dryrun")
        self.assertTrue("util_RIFT_pseudo_pipe.py" in result.output)

    def test_make_ini(self):
        """Check that a bayeswave config file can be built."""
        apply_page(file = "https://git.ligo.org/asimov/data/-/raw/main/defaults/production-pe.yaml", event=None, ledger=self.ledger)
        apply_page(file = "https://git.ligo.org/asimov/data/-/raw/main/defaults/production-pe-priors.yaml", event=None, ledger=self.ledger)
        event = "GW150914_095045"
        pipeline = "bayeswave"
        apply_page(file = f"https://git.ligo.org/asimov/data/-/raw/main/tests/{event}.yaml", event=None, ledger=self.ledger)
        apply_page(file = f"https://git.ligo.org/asimov/data/-/raw/main/tests/{pipeline}.yaml", event=event, ledger=self.ledger)

        f = io.StringIO()
        with contextlib.redirect_stdout(f):
            production = self.ledger.get_event(event)[0].productions[0]
            with set_directory(os.path.join("checkouts", event, config.get("general", "calibration_directory"))):
                production.make_config(f"{production.name}.ini")
            self.assertTrue(os.path.exists(os.path.join(config.get("project", "root"),
                                                        "checkouts",
                                                        event,
                                                        config.get("general", "calibration_directory"),
                                                        f"{production.name}.ini")))

        
    def test_build_api(self):
        """Check that a bayeswave DAG can be built."""
        apply_page(file = "https://git.ligo.org/asimov/data/-/raw/main/defaults/production-pe.yaml", event=None, ledger=self.ledger)
        apply_page(file = "https://git.ligo.org/asimov/data/-/raw/main/defaults/production-pe-priors.yaml", event=None, ledger=self.ledger)
        event = "GW150914_095045"
        pipeline = "bayeswave"
        apply_page(file = f"https://git.ligo.org/asimov/data/-/raw/main/tests/{event}.yaml", event=None, ledger=self.ledger)
        apply_page(file = f"https://git.ligo.org/asimov/data/-/raw/main/tests/{pipeline}.yaml", event=event, ledger=self.ledger)

        f = io.StringIO()
        with contextlib.redirect_stdout(f):
            production = self.ledger.get_event(event)[0].productions[0]
            with set_directory(os.path.join("checkouts", event, config.get("general", "calibration_directory"))):
                production.make_config(f"{production.name}.ini")
            production.pipeline.build_dag(dryrun=True)
            self.assertTrue("bayeswave_pipe" in f.getvalue())

#    @unittest.skipIf(not os.path.exists(os.path.join(config.get("pipelines", "environment"), "bin", "bayeswave_pipe")),
#                    "Bayeswave Pipe isnt installed on the test system")
    @patch('subprocess.Popen')
    def test_submit_api(self, mock_popen):
        """Check that a bayeswave config file can be submitted."""
        apply_page(file = "https://git.ligo.org/asimov/data/-/raw/main/defaults/production-pe.yaml", event=None, ledger=self.ledger)
        apply_page(file = "https://git.ligo.org/asimov/data/-/raw/main/defaults/production-pe-priors.yaml", event=None, ledger=self.ledger)
        event = "GW150914_095045"
        pipeline = "bayeswave"
        apply_page(file = f"https://git.ligo.org/asimov/data/-/raw/main/tests/{event}.yaml", event=None, ledger=self.ledger)
        apply_page(file = f"https://git.ligo.org/asimov/data/-/raw/main/tests/{pipeline}.yaml", event=event, ledger=self.ledger)
        f = io.StringIO()
        with contextlib.redirect_stdout(f):
            production = self.ledger.get_event(event)[0].productions[0]
            with set_directory(os.path.join("checkouts", event, config.get("general", "calibration_directory"))):
                production.make_config(f"{production.name}.ini")
            # We need to make the workdir as this ought to be done by bayeswave_pipe
            os.makedirs(os.path.join("working", event, production.name))

            mock_popen.returncode=0
            mock_popen.return_value.communicate.return_value=(b"Blah blah blah To submit: just run this", b"Lots of stuff on stderr")

            production.pipeline.build_dag(dryrun=False)

        with contextlib.redirect_stdout(f):

            mock_popen.returncode=0
            mock_popen.return_value.communicate.return_value=(b"submitted to cluster 999", b"Lots of stuff on stderr")

            
            production.pipeline.submit_dag(dryrun=False)
            self.ledger.update_event(production.event)
        
        self.assertEqual(production.job_id, 999)
        self.assertEqual(self.ledger.get_event(event)[0].productions[0].job_id, 999)

    def test_presubmit_mocked(self):
        """Check that a bayeswave submit file should be altered"""
        apply_page(file = "https://git.ligo.org/asimov/data/-/raw/main/defaults/production-pe.yaml", event=None, ledger=self.ledger)
        apply_page(file = "https://git.ligo.org/asimov/data/-/raw/main/defaults/production-pe-priors.yaml", event=None, ledger=self.ledger)
        event = "GW150914_095045"
        pipeline = "bayeswave"
        apply_page(file = f"https://git.ligo.org/asimov/data/-/raw/main/tests/{event}.yaml", event=None, ledger=self.ledger)
        apply_page(file = f"https://git.ligo.org/asimov/data/-/raw/main/tests/{pipeline}.yaml", event=event, ledger=self.ledger)

        f = io.StringIO()
        with contextlib.redirect_stdout(f):
            production = self.ledger.get_event(event)[0].productions[0]
            with set_directory(os.path.join("checkouts", event, config.get("general", "calibration_directory"))):
                production.make_config(f"{production.name}.ini")
            production.pipeline.build_dag(dryrun=True)
            self.assertTrue("bayeswave_pipe" in f.getvalue())
        # We need to make the workdir as this ought to be done by bayeswave_pipe
        os.makedirs(os.path.join("working", event, production.name))
        with set_directory(os.path.join(config.get("general", "rundir_default"), event, production.name)):
            with open("bayeswave_post.sub", "w") as submit_file:
                submit_file.write("This is some test text and is just garbage")

        production.pipeline.before_submit()

        with set_directory(os.path.join(config.get("general", "rundir_default"), event, production.name)):
            with open("bayeswave_post.sub", "r") as submit_file:
                self.assertTrue("request_disk" in submit_file.read())

    @patch('subprocess.Popen.communicate')
    @patch('subprocess.Popen')
    def test_bad_dag_build(self, mock_popen, mock_pcomm):
        """Check that things behave as expected if the DAG file can't be build."""
        apply_page(file = "https://git.ligo.org/asimov/data/-/raw/main/defaults/production-pe.yaml", event=None, ledger=self.ledger)
        apply_page(file = "https://git.ligo.org/asimov/data/-/raw/main/defaults/production-pe-priors.yaml", event=None, ledger=self.ledger)
        event = "GW150914_095045"
        pipeline = "bayeswave"
        apply_page(file = f"https://git.ligo.org/asimov/data/-/raw/main/tests/{event}.yaml", event=None, ledger=self.ledger)
        apply_page(file = f"https://git.ligo.org/asimov/data/-/raw/main/tests/{pipeline}.yaml", event=event, ledger=self.ledger)

        mock_popen.returncode=1 #"Could not build the DAG file"
        mock_popen.return_value.communicate.return_value=(b"Could not be created", b"Lots of stuff on stderr")
        
        production = self.ledger.get_event(event)[0].productions[0]
        with set_directory(os.path.join("checkouts", event, config.get("general", "calibration_directory"))):
            production.make_config(f"{production.name}.ini")
        
        with self.assertRaises(PipelineException):
            production.pipeline.build_dag(dryrun=False)



"""BayesWave Pipeline specification."""

import configparser
import glob
import os
import re
import subprocess
from shutil import copyfile, copytree

import numpy as np

from asimov import config
from asimov.utils import set_directory

from ..git import AsimovFileNotFound
from ..pipeline import Pipeline, PipelineException, PipelineLogger
from ..storage import AlreadyPresentException, Store


class BayesWave(Pipeline):
    """
    The BayesWave Pipeline.

    Parameters
    ----------
    production : :class:`asimov.Production`
       The production object.
    category : str, optional
        The category of the job.
        Defaults to "C01_offline".
    """

    name = "BayesWave"
    STATUS = {"wait", "stuck", "stopped", "running", "finished"}

    def __init__(self, production, category=None):
        super(BayesWave, self).__init__(production, category)
        self.logger.info("Using the Bayeswave pipeline")
        if not production.pipeline.lower() == "bayeswave":
            raise PipelineException

        try:
            self.category = config.get("general", "calibration_directory")
        except configparser.NoOptionError:
            self.category = "C01_offline"
            self.logger.info("Assuming C01_offline calibration.")

        if not production.meta.get("quality", {}).get("lowest minimum frequency", None):
            production.meta["quality"]["lowest minimum frequency"] = self.flow

    def build_dag(self, user=None, dryrun=False):
        """
        Construct a DAG file in order to submit a production to the
        condor scheduler using bayeswave_pipe

        Parameters
        ----------
        production : str
           The production name.
        user : str
           The user accounting tag which should be used to run the job.

        Raises
        ------
        PipelineException
           Raised if the construction of the DAG fails.
        """
        if self.production.event.repository:
            try:
                gps_file = self.production.get_timefile()
            except AsimovFileNotFound:
                if "event time" in self.production.meta:
                    gps_time = self.production.get_meta("event time")
                    with set_directory(
                        os.path.join(
                            self.production.event.repository.directory, self.category
                        )
                    ):
                        with open("gpstime.txt", "w") as f:
                            f.write(str(gps_time))
                            gps_file = os.path.join(
                                f"{self.production.category}", "gpstime.txt"
                            )
                            self.production.event.repository.add_file(
                                "gpstime.txt", gps_file
                            )
                else:
                    raise PipelineException("Cannot find the event time.")
        else:
            gps_time = self.production.get_meta("event time")
            with open("gpstime.txt", "w") as f:
                f.write(str(gps_time))
                gps_file = os.path.join("gpstime.txt")

        if self.production.event.repository:
            ini = self.production.get_configuration()
            if not user:
                if self.production.get_meta("user"):
                    user = self.production.get_meta("user")
                else:
                    user = ini._get_user()

            ini.update_accounting(user)

            if "queue" in self.production.meta:
                queue = self.production.meta["queue"]
            else:
                queue = "Priority_PE"

            ini.set_queue(queue)

            ini.save()

            ini = ini.ini_loc

        else:
            ini = f"{self.production.name}.ini"

        if self.production.rundir:
            rundir = self.production.rundir
        else:
            rundir = os.path.join(
                config.get("general", "rundir_default"),
                self.production.event.name,
                self.production.name,
            )
            self.production.rundir = rundir

        gps_time = self.production.get_meta("event time")

        pipe_cmd = os.path.join(
            config.get("pipelines", "environment"), "bin", "bayeswave_pipe"
        )

        command = [
            pipe_cmd,
            # "-l", f"{gps_file}",
            f"--trigger-time={gps_time}",
        ]

        if "cache files" in self.production.meta["data"]:
            if len(self.production.meta["data"]["cache files"]) > 0:
                # Skip the datafind step if the data is already provided.
                command += ["--skip-datafind"]

        if "copy frames" in self.production.meta["scheduler"]:
            if self.production.meta["scheduler"]["copy frames"]:
                command += ["--copy-frames"]

        if "osg" in self.production.meta["scheduler"]:
            if self.production.meta["scheduler"]["osg"]:
                command += ["--transfer-files"]

                if "copy frames" not in self.production.meta["scheduler"]:
                    command += ["--osg-deploy"]
                if "copy frames" in self.production.meta["scheduler"]:
                    if not self.production.meta["scheduler"]["copy frames"]:
                        command += ["--osg-deploy"]

        command += [
            "-r",
            self.production.rundir,
            ini,
        ]

        self.logger.info(" ".join(command))

        if dryrun:
            print(" ".join(command))
            self.logger.info(" ".join(command))
        else:

            pipe = subprocess.Popen(
                command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
            )
            out, err = pipe.communicate()
            if "To submit:" not in str(out):
                self.production.status = "stuck"
                self.logger.error("Could not create a DAG file")
                self.logger.info(f"{command}")
                self.logger.debug(out)
                self.logger.debug(err)
                raise PipelineException("The DAG file could not be created.")
            else:
                self.logger.info("DAG file created")
                self.logger.debug(out)

    def detect_completion(self):
        psds = self.collect_assets()["psds"]
        if len(list(psds.values())) > 0:
            return True
        else:
            self.logger.info("Bayeswave job completion was not detected.")
            return False

    def _convert_psd(self, ascii_format, ifo):
        """
        Convert an ascii format PSD to XML.

        Parameters
        ----------
        ascii_format : str
           The location of the ascii format file.
        ifo : str
           The IFO which this PSD is for.
        """
        command = [
            "convert_psd_ascii2xml",
            "--fname-psd-ascii",
            f"{ascii_format}",
            "--conventional-postfix",
            "--ifo",
            f"{ifo}",
        ]
        self.logger.info(" ".join(command))
        pipe = subprocess.Popen(
            command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
        )
        out, err = pipe.communicate()

        if err:
            self.production.status = "stuck"
            if hasattr(self.production.event, "issue_object"):
                raise Exception(
                    f"An XML format PSD could not be created.\n{command}\n{out}\n\n{err}",
                )
            else:
                raise Exception(
                    f"An XML format PSD could not be created.\n{command}\n{out}\n\n{err} ",
                )
        else:
            asset = f"{ifo.upper()}-psd.xml.gz"
            git_location = os.path.join(
                self.production.category,
                "psds",
                f"{self.production.meta['likelihood']['sample rate']}",
            )
            self.production.event.repository.add_file(
                asset,
                os.path.join(git_location, f"{ifo}-psd.xml.gz"),
                commit_message=f"Added the xml format PSD for {ifo}.",
            )

    def after_completion(self):

        try:
            for ifo, psd in self.collect_assets()["psds"].items():
                self._convert_psd(ascii_format=psd, ifo=ifo)
        except Exception as e:
            self.logger.error("Failed to convert the PSDs to XML")
            self.logger.exception(e)

        try:
            self.collect_pages()
        except FileNotFoundError:
            PipelineLogger(
                message=b"Failed to copy megaplot pages.",
                production=self.production.name,
            )

        try:
            self.collect_assets()
            self.store_assets()
        except Exception:
            PipelineLogger(
                message=b"Failed to store PSDs.",
                issue=self.production.event.issue_object,
                production=self.production.name,
            )

        if "supress" in self.production.meta["quality"]:
            for ifo in self.production.meta["quality"]["supress"]:
                if ifo in self.production.meta["interferometers"]:
                    self.supress_psd(
                        ifo,
                        self.production.meta["quality"]["supress"][ifo]["lower"],
                        self.production.meta["quality"]["supress"][ifo]["upper"],
                    )

        self.production.meta.update(self.collect_assets())

        self.production.status = "uploaded"

    @property
    def flow(self):
        """
        Calculate the lower frequency for the bayeswave job.

        This is required as bayeswave needs to be passed the lowest
        minimum frequency from the list of interferometer
        lower frequencies.
        """
        return min(self.production.meta["quality"]["minimum frequency"].values())

    def before_submit(self):
        """
        Horribly hack the sub files to add `request_disk`
        """
        sub_files = glob.glob(f"{self.production.rundir}/*.sub")
        for sub_file in sub_files:
            with open(sub_file, "r") as f_handle:
                original = f_handle.read()
            with open(sub_file, "w") as f_handle:
                self.logger.info(f"Adding request_disk = {64000} to {sub_file}")
                f_handle.write(f"request_disk = {64000}\n" + original)
        python_files = glob.glob(f"{self.production.rundir}/*.py")
        for py_file in python_files:
            with open(py_file, "r") as f_handle:
                original = f_handle.read()
            with open(py_file, "w") as f_handle:
                self.logger.info("Fixing shebang")
                path = os.path.join(
                    config.get("pipelines", "environment"), "bin", "python"
                )
                f_handle.write(f"#! {path}\n" + original)

    def submit_dag(self, dryrun=False):
        """
        Submit a DAG file to the condor cluster.

        Parameters
        ----------
        dryrun: bool
           If True then the DAG will not be submitted but all of the
           commands will be printed to stdout.

        Returns
        -------
        int
           The cluster ID assigned to the running DAG file.
        PipelineLogger
           The pipeline logger message.

        Raises
        ------
        PipelineException
           This will be raised if the pipeline fails to submit the job.
        """
        self.before_submit()

        command = [
            "condor_submit_dag",
            "-batch-name",
            f"bwave/{self.production.event.name}/{self.production.name}",
            f"{self.production.name}.dag",
        ]

        if dryrun:
            print(" ".join(command))

        else:
            with set_directory(self.production.rundir):
                try:
                    dagman = subprocess.Popen(
                        command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
                    )
                except FileNotFoundError as e:
                    self.logger.exception(e)
                    raise PipelineException(
                        "It looks like condor isn't installed on this system.\n"
                        f"""I wanted to run {" ".join(command)}."""
                    ) from e

                stdout, stderr = dagman.communicate()

            if "submitted to cluster" in str(stdout):
                cluster = re.search(
                    r"submitted to cluster ([\d]+)", str(stdout)
                ).groups()[0]
                self.production.status = "running"
                self.production.job_id = int(cluster)
                self.logger.info(
                    f"Successfully submitted to cluster {self.production.job_id}"
                )
                self.logger.debug(stdout)
                return (int(cluster),)
            else:
                self.logger.info(stdout)
                self.logger.error(stderr)
                raise PipelineException(
                    f"The DAG file could not be submitted.\n\n{stdout}\n\n{stderr}",
                )

    def upload_assets(self):
        """
        Upload the PSDs from this job.
        """
        sample = self.production.meta["likelihood"]["sample rate"]
        git_location = os.path.join(self.category, "psds")

        for detector, asset in self.collect_assets()["psds"]:
            self.production.event.repository.add_file(
                asset,
                os.path.join(git_location, str(sample), f"{detector}-psd.dat"),
                commit_message=f"Added the PSD for {detector}.",
            )

    def store_assets(self):
        """
        Add the assets to the store.
        """

        sample_rate = self.production.meta["likelihood"]["sample rate"]
        self.logger.info(self.collect_assets())
        for detector, asset in self.collect_assets()["psds"].items():
            store = Store(root=config.get("storage", "directory"))
            try:
                store.add_file(
                    self.production.event.name,
                    self.production.name,
                    file=asset,
                    new_name=f"{detector}-{sample_rate}-psd.dat",
                )
            except Exception as e:
                self.logger.error(
                    f"There was a problem committing the PSD for {detector} to the store."
                )
                self.logger.exception(e)

    def collect_logs(self):
        """
        Collect all of the log files which have been produced by this production and
        return their contents as a dictionary.
        """
        messages = {}

        logfile = os.path.join(
            config.get("logging", "directory"),
            self.production.event.name,
            self.production.name,
            "asimov.log",
        )
        with open(logfile, "r") as log_f:
            message = log_f.read()
            messages["production"] = message

        logs = glob.glob(f"{self.production.rundir}/logs/*.err") + glob.glob(
            f"{self.production.rundir}/*.err"
        )
        for log in logs:
            with open(log, "r") as log_f:
                message = log_f.read()
                messages[log.split("/")[-1]] = message
        return messages

    def collect_assets(self):
        """
        Collect the assets for this job and commit them to the event repository.
        Since this job also generates the PSDs these should be added to the production ledger.
        """
        psds = {}
        results_dir = glob.glob(f"{self.production.rundir}/trigtime_*")[0]
        for det in self.production.meta["interferometers"]:
            asset = os.path.join(
                results_dir, "post", "clean", f"glitch_median_PSD_forLI_{det}.dat"
            )
            if os.path.exists(asset):
                psds[det] = asset

        outputs = {}
        outputs["psds"] = psds

        xml_psds = {}
        for det in self.production.meta["interferometers"]:
            asset = os.path.join(
                self.production.event.repository.directory,
                self.production.category,
                "psds",
                f"{self.production.meta['likelihood']['sample rate']}",
                f"{det.upper()}-psd.xml.gz",
            )
            if os.path.exists(asset):
                xml_psds[det] = os.path.abspath(asset)

        outputs["xml psds"] = xml_psds

        return outputs

    def supress_psd(self, ifo, fmin, fmax):
        """
        Suppress portions of a PSD.
        Author: Carl-Johan Haster - August 2020
        (Updated for asimov by Daniel Williams - November 2020
        """
        store = Store(root=config.get("storage", "directory"))
        sample_rate = self.production.meta["quality"]["sample-rate"]
        orig_PSD_file = np.genfromtxt(
            os.path.join(
                self.production.event.repository.directory,
                self.category,
                "psds",
                str(sample_rate),
                f"{ifo}-psd.dat",
            )
        )

        self.logger.info("PSD supression has been set")
        self.logger.info(
            f"{ifo}-psd.dat will be supressed between {fmin}-Hz and {fmax}-Hz"
        )

        freq = orig_PSD_file[:, 0]
        PSD = orig_PSD_file[:, 1]

        suppression_region = np.logical_and(
            np.greater_equal(freq, fmin), np.less_equal(freq, fmax)
        )

        # Suppress the PSD in this region

        PSD[suppression_region] = 1.0

        new_PSD = np.vstack([freq, PSD]).T

        asset = f"{ifo}-psd.dat"
        np.savetxt(asset, new_PSD, fmt="%+.5e")

        destination = os.path.join(
            self.category, "psds", str(sample_rate), f"{ifo}-psd.dat"
        )

        self.logger.info(
            f"{ifo}-psd.dat has been supressed between {fmin}-Hz and {fmax}-Hz"
        )

        try:
            self.production.event.repository.add_file(
                asset, destination
            )  # , message=f"Added the supresed {ifo} PSD")
        except Exception as e:
            self.logger.error(
                "The supressed PSD could not be committed to the repository"
            )
            self.logger.exception(e)

        copyfile(asset, f"{ifo}-{sample_rate}-psd-suppresed.dat")
        try:
            store.add_file(
                self.production.event.name,
                self.production.name,
                file=f"{ifo}-{sample_rate}-psd-suppresed.dat",
            )
        except AlreadyPresentException:
            self.logger.warning(
                "Attempted to add a supressed PSD which already exists."
            )

    def resurrect(self):
        """
        Attempt to ressurrect a failed job.
        """
        count = len(glob.glob(os.path.join(self.production.rundir, "*.dag.rescue*")))

        if (count < 5) and (count > 0):
            self.submit_dag()
            self.logger.info(f"Bayeswave job was resurrected for the {count} time.")
        else:
            self.logger.error(
                "Bayeswave resurrection not completed as there have already been 5 attempts"
            )

    def html(self):
        """Return the HTML representation of this pipeline."""
        pages_dir = os.path.join(self.production.event.name, self.production.name)
        out = ""
        if self.production.status in {"finished", "uploaded"}:
            out += """<div class="asimov-pipeline">"""
            out += (
                f"""<p><a href="{pages_dir}/index.html">Full Megaplot output</a></p>"""
            )
            out += f"""<img height=200 src="{pages_dir}/plots/clean_whitened_residual_histograms.png"</src>"""

            out += """</div>"""

        return out

    def collect_pages(self):
        """Collect the HTML output of the pipeline."""
        results_dir = glob.glob(f"{self.production.rundir}/trigtime_*")[0]
        pages_dir = os.path.join(
            config.get("general", "webroot"),
            self.production.event.name,
            self.production.name,
        )
        os.makedirs(pages_dir, exist_ok=True)
        copyfile(
            os.path.join(results_dir, "index.html"),
            os.path.join(pages_dir, "index.html"),
        )
        copytree(
            os.path.join(results_dir, "html"),
            os.path.join(pages_dir, "html"),
            dirs_exist_ok=True,
        )
        copytree(
            os.path.join(results_dir, "plots"),
            os.path.join(pages_dir, "plots"),
            dirs_exist_ok=True,
        )



"""
This module contains tools to log data from runs.
Unlike a conventional python process, processes run by
asimov may wish to send information to a number of different
locations, and store information for review purposes.
"""

import logging
import datetime
import asimov.database as database
from asimov import mattermost

# from .gitlab import EventIssue


logging.getLogger("werkzeug").setLevel(logging.WARNING)
logging.getLogger("MARKDOWN").setLevel(logging.WARNING)
logging.getLogger("matplotlib").setLevel(logging.WARNING)
logging.getLogger("git").setLevel(logging.WARNING)


class AsimovLogger:
    def __init__(self, logfile):
        """
        An object for logging asimov productions.

        Parameters
        ----------
        logfile : str
           The location of the log file to be written.
        """
        self.logfile = logfile
        self.mattermost = mattermost.Mattermost()
        self.file_logger = logging.basicConfig(filename=logfile)

        logger_name = "asimov"
        self.logger = logging.getLogger(logger_name)

        formatter = logging.Formatter(
            "%(asctime)s [%(name)s][%(levelname)s] %(message)s", "%Y-%m-%d %H:%M:%S"
        )
        print_formatter = logging.Formatter(
            "[%(levelname)s] %(message)s", "%Y-%m-%d %H:%M:%S"
        )

        ch = logging.StreamHandler()
        ch.setFormatter(print_formatter)
        ch.setLevel(logging.ERROR)

        fh = logging.FileHandler(logfile)
        fh.setFormatter(formatter)
        fh.setLevel(logging.INFO)

        self.logger.addHandler(ch)
        self.logger.addHandler(fh)

    def log(
        self,
        level,
        message,
        event=None,
        production=None,
        pipeline=None,
        module=None,
        time=None,
    ):
        """
        Record a log message.

        Parameters
        ----------
        level : str, {DEBUG, INFO, WARNING, ERROR}
           The level of the logging message.
        event : ``asimov.event.Event``
           The event which this message is for.
        production : ``asimov.event.Production``, optional
           The production this message is for.
        message : str
           The free-form message for the log message.
        channels : list, or str, {"file", "mattermost", "gitlab"}, optional
           The list of places where the log message should be sent.
           Defaults to file only.
        """

        logger_levels = {
            "debug": logging.DEBUG,
            "info": logging.INFO,
            "warning": logging.WARNING,
            "error": logging.ERROR,
            "update": 9,
        }

        if not production:
            production = ""

        self.logger.log(logger_levels[level.lower()], message)

    def info(self, message, production=None, channels="file"):
        """
        Record an information message.

        Parameters
        ----------

        production : ``asimov.event.Production``, optional
           The production this message is for.
        message : str
           The free-form message for the log message.
        channels : list, or str, {"file", "mattermost", "gitlab"}, optional
           The list of places where the log message should be sent.
           Defaults to file only.
        """

        self.log("info", message, production, channels)

    def error(self, message, production=None, channels="file"):
        """
        Record an error message.

        Parameters
        ----------

        production : ``asimov.event.Production``, optional
           The production this message is for.
        message : str
           The free-form message for the log message.
        channels : list, or str, {"file", "mattermost", "gitlab"}, optional
           The list of places where the log message should be sent.
           Defaults to file only.
        """

        self.log("error", message, production, channels)

    def warning(self, message, production=None, channels="file"):
        """
        Record an warning message.

        Parameters
        ----------

        production : ``asimov.event.Production``, optional
           The production this message is for.
        message : str
           The free-form message for the log message.
        channels : list, or str, {"file", "mattermost", "gitlab"}, optional
           The list of places where the log message should be sent.
           Defaults to file only.
        """

        self.log("warning", message, production, channels)

    def list(self, offset=0, length=10, **filters):
        class LogEntry:
            def __init__(self, data):
                data = data.strip().split(":")
                self.level = data[0].lower()
                try:
                    self.module = data[1]
                    if self.module == "root":
                        self.module = "asimov"
                except KeyError:
                    self.module = None
                try:
                    self.message = data[2]
                except KeyError:
                    self.message = None

            def __dictrepr__(self):
                return dict(level=self.level, module=self.module, message=self.message)

        entries = []
        with open(self.logfile, "r") as filehandle:
            for line in filehandle:
                entries.append(LogEntry(line))

        def apply_filter(entries, parameter, value):
            entries = filter(lambda x: getattr(x, parameter) == value, entries)
            return entries

        if filters:
            for parameter, value in filters.items():
                entries = apply_filter(entries, parameter, value)
            entries = list(entries)

        if offset:
            return entries[-int(length) + int(offset): -int(offset)]
        else:
            return entries[-int(length):]


class DatabaseLogger(AsimovLogger):
    """
    This class implements a database-backed log collector.
    """

    def __init__(self):
        self.database = database.Logger

    def log(
        self, level, message, event=None, production=None, pipeline=None, module=None
    ):
        """
        Record a log message.

        Parameters
        ----------
        level : str, {DEBUG, INFO, WARNING, ERROR}
           The level of the logging message.
        production : ``asimov.event.Production``, optional
           The production this message is for.
        message : str
           The free-form message for the log message.
        channels : list, or str, {"file", "mattermost", "gitlab"}, optional
           The list of places where the log message should be sent.
           Defaults to file only.
        """

        entry = self.database(
            pipeline, module, level, datetime.datetime.now(), event, production, message
        )
        entry.save()

    def list(self, **filters):
        """
        Return all logging events which satisfy a set of ``filters``.

        Examples
        --------

        """

        events = self.database.list(**filters)
        return events



"""
Code for the project ledger.
"""
import yaml

import os
import shutil

import asimov
import asimov.database
from asimov import config
from asimov.event import Event, Production
from asimov.utils import update, set_directory


class Ledger:
    @classmethod
    def create(cls, name=None, engine=None, location=None):
        """
        Create a ledger.
        """

        if not engine:
            engine = config.get("ledger", "engine")

        if engine == "yamlfile":
            YAMLLedger.create(location=location, name=name)

        elif engine in {"tinydb", "mongodb"}:
            DatabaseLedger.create()
        elif engine == "gitlab":
            raise NotImplementedError(
                "This hasn't been ported to the new interface yet. Stay tuned!"
            )

        elif engine == "gitlab":
            raise NotImplementedError(
                "This hasn't been ported to the new interface yet. Stay tuned!"
            )


class YAMLLedger(Ledger):
    def __init__(self, location=None):
        if not location:
            location = os.path.join(".asimov", "ledger.yml")
        self.location = location
        with open(location, "r") as ledger_file:
            self.data = yaml.safe_load(ledger_file)

        self.data["events"] = [
            update(self.get_defaults(), event, inplace=False)
            for event in self.data["events"]
        ]
        self.events = {ev["name"]: ev for ev in self.data["events"]}
        self._all_events = [
            Event(**self.events[event], ledger=self)
            for event in self.events.keys()
            ]
        self.data.pop("events")

    @classmethod
    def create(cls, name, location=None):
        if not location:
            location = os.path.join(".asimov", "ledger.yml")
        data = {}
        data["asimov"] = {}
        data["asimov"]["version"] = asimov.__version__
        data["events"] = []
        data["project"] = {}
        data["project"]["name"] = name
        with open(location, "w") as ledger_file:
            ledger_file.write(yaml.dump(data, default_flow_style=False))

    def update_event(self, event):
        """
        Update an event in the ledger with a changed event object.
        """
        self.events[event.name] = event.to_dict()
        self.save()

    def delete_event(self, event_name):
        """
        Remove an event from the ledger.

        Parameters
        ----------
        event_name : str
           The name of the event to remove from the ledger.
        """
        event = self.events.pop(event_name)
        if "trash" not in self.data:
            self.data["trash"] = {}
        if "events" not in self.data["trash"]:
            self.data["trash"]["events"] = {}
        self.data["trash"]["events"][event_name] = event
        self.save()

    def save(self):
        """
        Update the ledger YAML file with the data from the various events.

        Notes
        -----
        The save function checks the difference between the default values for each production and event
        before saving them, in order to attempt to reduce the duplication within the ledger.


        """
        self.data["events"] = list(self.events.values())
        with set_directory(config.get("project", "root")):
            # First produce a backup of the ledger
            shutil.copy(self.location, self.location + ".bak")
            with open(self.location + "_tmp", "w") as ledger_file:
                ledger_file.write(yaml.dump(self.data, default_flow_style=False))
                ledger_file.flush()
                # os.fsync(ledger_file.fileno())
            os.replace(self.location + "_tmp", self.location)

    def add_event(self, event):
        if "events" not in self.data:
            self.data["events"] = []

        self.events[event.name] = event.to_dict()
        self.save()

    def add_production(self, event, production):
        event.add_production(production)
        self.events[event.name] = event.to_dict()
        self.save()

    def get_defaults(self):
        """
        Gather project-level defaults from the ledger.

        At present data, quality, priors, and likelihood settings can all be set at a project level as defaults.
        """
        defaults = {}
        if "data" in self.data:
            defaults["data"] = self.data["data"]
        if "priors" in self.data:
            defaults["priors"] = self.data["priors"]
        if "quality" in self.data:
            defaults["quality"] = self.data["quality"]
        if "likelihood" in self.data:
            defaults["likelihood"] = self.data["likelihood"]
        if "scheduler" in self.data:
            defaults["scheduler"] = self.data["scheduler"]
        return defaults

    def get_event(self, event=None):
        if event:
            return [Event(**self.events[event], ledger=self)]
        else:
            return self._all_events

    def get_productions(self, event=None, filters=None):
        """Get a list of productions either for a single event or for all events.

        Parameters
        ----------
        event : str
           The name of the event to pull productions from.
           Optional; if no event is specified then all of the productions are
           returned.

        filters : dict
           A dictionary of parameters to filter on.

        Examples
        --------
        FIXME: Add docs.

        """

        if event:
            productions = self.get_event(event).productions
        else:
            productions = []
            for event_i in self.get_event():
                for production in event_i.productions:
                    productions.append(production)

        def apply_filter(productions, parameter, value):
            productions = filter(
                lambda x: x.meta[parameter] == value
                if (parameter in x.meta)
                else (
                    getattr(x, parameter) == value if hasattr(x, parameter) else False
                ),
                productions,
            )
            return productions

        if filters:
            for parameter, value in filters.items():
                productions = apply_filter(productions, parameter, value)
        return list(productions)


class DatabaseLedger(Ledger):
    """
    Use a document database to store the ledger.
    """

    def __init__(self):
        if config.get("ledger", "engine") == "tinydb":
            self.db = asimov.database.AsimovTinyDatabase()
        else:
            self.db = asimov.database.AsimovTinyDatabase()

    @classmethod
    def create(cls):
        ledger = cls()
        ledger.db._create()
        return ledger

    def _insert(self, payload):
        """
        Store the payload in the correct database table.
        """

        if isinstance(payload, Event):
            id_number = self.db.insert("event", payload.to_dict(productions=False))
        elif isinstance(payload, Production):
            id_number = self.db.insert("production", payload.to_dict(event=False))

        return id_number

    @property
    def events(self):
        """
        Return all of the events in the ledger.
        """
        return [Event.from_dict(page) for page in self.db.tables["event"].all()]

    def get_defaults(self):
        raise NotImplementedError

    def get_event(self, event=None):
        """
        Find a specific event in the ledger and return it.
        """
        event_dict = self.db.query("event", "name", event)[0]
        return Event.from_dict(event_dict)

    def get_productions(self, event=None, filters=None):
        """Get a list of productions either for a single event or for all events.

        Parameters
        ----------
        event : str
           The name of the event to pull productions from.
           Optional; if no event is specified then all of the productions are
           returned.

        filters : dict
           A dictionary of parameters to filter on.

        Examples
        --------
        FIXME: Add docs.

        """

        if event:
            productions = self.get_event(event).productions
        else:
            productions = []
            for event_i in self.get_event():
                for production in event_i.productions:
                    productions.append(production)

        def apply_filter(productions, parameter, value):
            productions = filter(
                lambda x: x.meta[parameter] == value
                if (parameter in x.meta)
                else (
                    getattr(x, parameter) == value if hasattr(x, parameter) else False
                ),
                productions,
            )
            return productions

        if filters:
            for parameter, value in filters.items():
                productions = apply_filter(productions, parameter, value)
        return list(productions)



import unittest
import os
import shutil
import git

from importlib import reload
import contextlib
import io
from unittest.mock import patch

from click.testing import CliRunner

import asimov
from asimov.event import Production
from asimov.cli import review, project
from asimov.ledger import YAMLLedger
import asimov.event
from asimov.cli.project import make_project
from asimov.cli.application import apply_page
from asimov.ledger import YAMLLedger

EVENTS = ["GW150914_095045", "GW190924_021846", "GW190929_012149", "GW191109_010717"]
pipelines = {"bayeswave"}

TEST_YAML = """
name: S000000xx
working directory: {0}/tests/tmp/
repository: {0}/tests/test_data/s000000xx/
data: 
  channels:
    L1: /this/is/fake
  calibration: 
    L1: Fake
interferometers: 
- L1
quality: {{}}
productions:
- Prod0:
    pipeline: lalinference
    comment: PSD production
    status: wait
    review:
    - message: Second.
      status: APPROVED
      timestamp: 2021-03-17 21:58:05
    - message: This is a review message.
      status: REJECTED
      timestamp: 2021-03-17 20:58:05
    - message: Third.
      timestamp: 2021-03-18 20:58:05
"""
TEST_YAML_2 = """
name: S000000xx
data: 
  channels:
    L1: /this/is/fake
  calibration: 
    L1: Fake
working directory: {0}/tests/tmp/
repository: {0}/tests/test_data/s000000xx/
interferometers: 
- L1
quality: {{}}
productions:
- Prod0:
    pipeline: lalinference
    comment: PSD production
    status: wait
"""
class ReviewTests(unittest.TestCase):
    """All tests of the YAML event format."""
    @classmethod
    def setUpClass(cls):
        cls.cwd = os.getcwd()
        git.Repo.init(cls.cwd+"/tests/test_data/s000000xx/")
        
    def setUp(self):
        os.makedirs(f"{self.cwd}/tests/tmp/project")
        os.chdir(f"{self.cwd}/tests/tmp/project")
        make_project(name="Test project", root=f"{self.cwd}/tests/tmp/project")
        self.ledger = YAMLLedger(f".asimov/ledger.yml")
        apply_page(file = "https://git.ligo.org/asimov/data/-/raw/main/defaults/production-pe.yaml", event=None, ledger=self.ledger)
        apply_page(file = "https://git.ligo.org/asimov/data/-/raw/main/events/gwtc-2-1/GW150914_095045.yaml", event=None, ledger=self.ledger)

        self.event = asimov.event.Event.from_yaml(TEST_YAML.format(self.cwd),
                                                  ledger=self.ledger)
        self.event_no_review = asimov.event.Event.from_yaml(TEST_YAML_2.format(self.cwd),
                                                            ledger=self.ledger)

    def tearDown(self):
        os.chdir(self.cwd)
        shutil.rmtree(f"{self.cwd}/tests/tmp/project")
    
    def test_review_parsing(self):
        """Check that review messages get parsed correctly."""
        self.assertEqual(self.event.productions[0].review[0].status, "REJECTED")

    def test_empty_register_creation(self):
        """Check that productions initialise an empty register if no review information attached."""
        self.assertEqual(len(self.event_no_review.productions[0].review), 0)

    def test_review_message_sort(self):
        """Check that messages are correctly sorted."""
        self.assertEqual(self.event.productions[0].review[-1].message, "Third.")
        self.assertEqual(self.event.productions[0].review[1].message, "Second.")

    def test_gets_latest_status(self):
        """Ensure latest status is reported."""
        self.assertEqual(self.event.productions[0].review.status, "APPROVED")
        
    def test_outputs_to_dict(self):
        """Check that dictionary output works"""
        self.assertEqual(self.event.productions[0].review.to_dicts()[0]['status'], 'REJECTED')


class ReviewCliTests(unittest.TestCase):
    """Test the CLI interfaces for review."""

    @classmethod
    def setUpClass(cls):
        cls.cwd = os.getcwd()

    def tearDown(self):
        os.chdir(self.cwd)
        shutil.rmtree(f"{self.cwd}/tests/tmp/")
    
    def setUp(self):
        os.makedirs(f"{self.cwd}/tests/tmp/project")
        os.chdir(f"{self.cwd}/tests/tmp/project")
        runner = CliRunner()
        result = runner.invoke(project.init,
                               ['Test Project', '--root', f"{self.cwd}/tests/tmp/project"])
        assert result.exit_code == 0
        assert result.output == '● New project created successfully!\n'
        self.ledger = YAMLLedger(".asimov/ledger.yml")

        f = io.StringIO()
        with contextlib.redirect_stdout(f):
            apply_page(file = "https://git.ligo.org/asimov/data/-/raw/main/defaults/production-pe.yaml", event=None, ledger=self.ledger)
            apply_page(file = "https://git.ligo.org/asimov/data/-/raw/main/defaults/production-pe-priors.yaml", event=None, ledger=self.ledger)
            for event in EVENTS:
                for pipeline in pipelines:
                    apply_page(file = f"https://git.ligo.org/asimov/data/-/raw/main/tests/{event}.yaml", event=None, ledger=self.ledger)
                    apply_page(file = f"https://git.ligo.org/asimov/data/-/raw/main/tests/{pipeline}.yaml", event=event, ledger=self.ledger)

    def test_add_review_to_event_reject(self):
        """Check that the CLI can add an event review"""
        with patch("asimov.current_ledger", new=YAMLLedger(".asimov/ledger.yml")):
            reload(asimov)
            reload(review)
            runner = CliRunner()

            for event in EVENTS:
                result = runner.invoke(review.review,
                                       ['add', event, "Prod0", "Rejected"])
                self.assertTrue(f"{event}/Prod0 rejected" in result.output)     

    def test_add_review_to_event_reject(self):
        """Check that the CLI can add an event review"""
        with patch("asimov.current_ledger", new=YAMLLedger(".asimov/ledger.yml")):
            reload(asimov)
            reload(review)
            runner = CliRunner()

            for event in EVENTS:
                result = runner.invoke(review.review,
                                       ['add', event, "Prod0", "Approved"])
                self.assertTrue(f"{event}/Prod0 approved" in result.output)          

    def test_add_review_to_event_unknown(self):
        """Check that the CLI rejects an unknown review status"""
        with patch("asimov.current_ledger", new=YAMLLedger(".asimov/ledger.yml")):
            reload(asimov)
            reload(review)
            runner = CliRunner()

            for event in EVENTS:
                result = runner.invoke(review.review,
                                       ['add', event, "Prod0", "Splorg"])
                self.assertTrue(f"Did not understand the review status splorg" in result.output)                          

    def test_show_review_no_review(self):
        """Check that the CLI can show a review report with no reviews"""
        with patch("asimov.current_ledger", new=YAMLLedger(".asimov/ledger.yml")):
            reload(asimov)
            reload(review)
            runner = CliRunner()

            for event in EVENTS:
                result = runner.invoke(review.review,
                                       ['status', event])
                self.assertTrue(f"No review information exists for this production." in result.output)            

    def test_show_review_with_review(self):
        """Check that the CLI can show a review report with no reviews"""
        with patch("asimov.current_ledger", new=YAMLLedger(".asimov/ledger.yml")):
            reload(asimov)
            reload(review)
            runner = CliRunner()

            for event in EVENTS:
                result = runner.invoke(review.review,
                                       ['add', event, "Prod0", "Approved"])                
                result = runner.invoke(review.review,
                                       ['status', event])
                self.assertTrue(f"approved" in result.output)       



"""
Asimov databse interface
------------------------

This module implements the asimov database and its interfaces.

Note that this approach to the database is a departure from what I set
up initially for the database backed logger, and I think it would be
better to use a document database going forward.

Implementation
--------------

Initially the implementation is designed to use TinyDB but that's
just because I want something I can work with easily on a laptop.
MongoDB would be a better long-term solution.

"""
from tinydb import Query, TinyDB

from asimov import config


class AsimovDatabase:
    pass


class AsimovTinyDatabase(AsimovDatabase):
    def __init__(self):

        database_path = config.get("ledger", "location")
        self.db = TinyDB(database_path)
        self.tables = {
            "event": self.db.table("event"),
            "production": self.db.table("production"),
        }
        self.Q = Query()

    @classmethod
    def _create(cls):
        db_object = cls()
        db_object.db.table("event")
        return db_object

    def insert(self, table, dictionary, doc_id=None):
        """
        Store an entire model in the dictionary.

        Parameters
        ----------
        table : str
           The name of the table which the document should be stored in.
        dictionary: dict
           The dicitonary containing the model.
        """
        doc_id = self.tables[table].insert(dictionary)
        return doc_id

    def query(self, table, parameter, value):
        pages = self.tables[table].search(Query()[parameter] == value)
        return pages



import unittest
import os
import shutil
from click.testing import CliRunner
from asimov.cli import project

class TestCLI_Projects(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cwd = os.getcwd()

    
    def setUp(self):
        os.makedirs(f"{self.cwd}/tests/tmp/project")
        os.chdir(f"{self.cwd}/tests/tmp/project")

    
    def tearDown(self):
        os.chdir(self.cwd)
        shutil.rmtree(f"{self.cwd}/tests/tmp/")
        
    def test_project_creation(self):
        os.chdir(f"{self.cwd}/tests/tmp/project")
        runner = CliRunner()
        result = runner.invoke(project.init,
                               ['Test Project', '--root', f"{self.cwd}/tests/tmp/project"])
        assert result.exit_code == 0
        assert result.output == '● New project created successfully!\n'

    def test_project_creation_fails_missing_name(self):
        """Check that the command fails if no name is provided"""
        os.chdir(f"{self.cwd}/tests/tmp/project")
        runner = CliRunner()
        result = runner.invoke(project.init,
                               [ '--root', f"{self.cwd}/tests/tmp/project"])
        assert result.exit_code == 2



"""
Make the event issues on the gitlab issue tracker from gracedb.
"""
import os

import asimov
from asimov.event import Event, DescriptionException
from asimov import config

from asimov import gitlab
from asimov import config

import numpy as np
import yaml

from git.exc import GitCommandError
import click

#from ligo.gracedb.rest import GraceDb, HTTPError 
#client = GraceDb(service_url=config.get("gracedb", "url"))
#r = client.ping()

#superevent_iterator = client.superevents('O3B_CBC_CATALOG')
#superevent_ids = [superevent['superevent_id'] for superevent in superevent_iterator]

server = gitlab.gitlab.Gitlab(config.get("gitlab", "url"), private_token=config.get("gitlab", "token"))
repository = server.projects.get(config.get("olivaw", "tracking_repository"))

CALIBRATION_NOTE = """
## Calibration envelopes
The following calibration envelopes have been found.
```yaml
---
{}
---
```
"""

@click.group()
def olivaw():
    """
    This is the main olivaw program which runs the DAGs for each event issue.
    """
    click.echo("Running olivaw")

    global rundir
    rundir = os.getcwd()


def find_calibrations(time):
    with open("LLO_calibs.txt") as llo_file:
        data_llo = llo_file.read().split("\n")
        data_llo = [datum for datum in data_llo if datum[-16:]=="FinalResults.txt"]
        times_llo = {int(datum.split("GPSTime_")[1].split("_C01")[0]): datum for datum in data_llo}
    
    with open("LHO_calibs.txt") as llo_file:
        data_lho = llo_file.read().split("\n")
        data_lho = [datum for datum in data_lho if datum[-16:]=="FinalResults.txt"]
        times_lho = {int(datum.split("GPSTime_")[1].split("_C01")[0]): datum for datum in data_lho}
        
    keys_llo = np.array(list(times_llo.keys())) 
    keys_lho = np.array(list(times_lho.keys())) 

    return {"H1": times_lho[keys_lho[np.argmin(np.abs(keys_lho - time))]], "L1": times_llo[keys_llo[np.argmin(np.abs(keys_llo - time))]], "V1": "/home/cbc/pe/O3/calibrationenvelopes/Virgo/V_O3a_calibrationUncertaintyEnvelope_magnitude5percent_phase35milliradians10microseconds.txt"}

    
# Update existing events
for event in gitlab_events:
    print(event.name)
    try:
        event.event_object._check_calibration()
    except DescriptionException:
        time = event.event_object.meta['event time'] 

        calibrations = find_calibrations(time)

        envelopes = yaml.dump({"calibration": calibrations})
        event.add_note(f"""
## Calibration envelopes
The following calibration envelopes have been found.
```yaml
---
{envelopes}
---
```
""")
>>>>>>> cb204b61f687395eb980468da3b8ced48c5c7e40

@click.option("--event", "event", default=None, help="The event which the ledger should be returned for, optional.")
@olivaw.command()
def calibration(event):    
    gitlab_events = gitlab.find_events(repository, subset=event)
    # Update existing events
    for event in gitlab_events:
        if "disable_repo" in event.event_object.meta:
            if event.event_object.meta['disable_repo'] == True:
                continue
        try:
            event.event_object._check_calibration()
        except DescriptionException:
            print(event.title)
            time = event.event_object.meta['event time'] 

            calibrations = find_calibrations(time)
            print(calibrations)
            # try:
            for ifo, envelope in calibrations.items():
                description = f"Added calibration {envelope} for {ifo}."
                try:
                    event.event_object.repository.add_file(os.path.join(f"/home/cal/public_html/uncertainty/O3C01/{ifo}", envelope), f"C01_offline/calibration/{ifo}.dat", 
                                                       commit_message=description)
                except GitCommandError as e:
                    if "nothing to commit," in e.stderr:
                        pass
                calibrations[ifo] = f"C01_offline/calibration/{ifo}.dat"
            envelopes = yaml.dump({"calibration": calibrations})
            event.add_note(CALIBRATION_NOTE.format(envelopes))

olivaw()



"""
Reporting functions
"""
from datetime import datetime

import os

import click
import pytz
import yaml
from pkg_resources import resource_filename

import otter
import otter.bootstrap as bt

from asimov import config, current_ledger

tz = pytz.timezone("Europe/London")


@click.group()
def report():
    """Produce reports about the state of the project."""
    pass


@click.option(
    "--location", "webdir", default=None, help="The place to save the report to"
)
@click.argument("event", default=None, required=False)
@report.command()
def html(event, webdir):
    """
    Return the ledger for a given event.
    If no event is specified then the entire production ledger is returned.
    """

    events = current_ledger.get_event(event)

    if not webdir:
        webdir = config.get("general", "webroot")

    report = otter.Otter(
        f"{webdir}/index.html",
        author="Asimov",
        title="Asimov project report",
        theme_location=resource_filename("asimov.cli", "report-theme"),
        config_file=os.path.join(".asimov", "asimov.conf"),
    )
    with report:

        style = """
<style>
        body {
        background-color: #f2f2f2;
        }

        .review-deprecated, .status-cancelled, .review-rejected {
        display: none;
        }

        .event-data {
        margin: 1rem;
        margin-bottom: 2rem;
        }

        .asimov-sidebar {
        position: sticky;
        top: 4rem;
        height: calc(100vh - 4rem);
        overflow-y: auto;
        }

        .asimov-analysis {
        padding: 1rem;
        background: lavenderblush;
        margin: 0.5rem;
        border-radius: 0.5rem;
        }

        .asimov-analysis-running, .asimov-analysis-processing {
        background: #DEEEFF;
        }

        .asimov-analysis-finished, .asimov-analysis-uploaded {
        background: #E1EDE4;
        }
        .asimov-analysis-stuck {
        background: #FFF5D9;
        }

        .asimov-status {
        float: right;
        clear: both;
        }
</style>
        """
        report + style

        script = """
<script type="text/javascript">
    window.onload = setupRefresh;

    function setupRefresh() {
      setTimeout("refreshPage();", 1000*60*15); // milliseconds
    }
    function refreshPage() {
       window.location = location.href;
    }
</script>
        """
        report + script
    with report:
        navbar = bt.Navbar(
            f"Asimov  |  {current_ledger.data['project']['name']}",
            background="navbar-dark bg-primary",
        )
        report + navbar

    events = sorted(events, key=lambda a: a.name)
    cards = "<div class='container-fluid'><div class='row'><div class='col-12 col-md-3 col-xl-2  asimov-sidebar'>"

    toc = """<nav><h6>Subjects</h6><ul class="list-unstyled">"""
    for event in events:
        toc += f"""<li><a href="#card-{event.name}">{event.name}</a></li>"""

    toc += "</ul></nav>"

    cards += toc
    cards += """</div><div class='events col-md-9 col-xl-10'
    data-isotope='{ "itemSelector": ".production-item", "layoutMode": "fitRows" }'>"""

    for event in events:
        card = ""
        # This is a quick test to try and improve readability
        card += event.html()

        # card += """<p class="card-text">Card text</p>""" #
        card += """
</div>
</div>"""
        cards += card

    cards += "</div></div>"
    with report:
        report + cards

    with report:
        time = f"Report generated at {datetime.now(tz):%Y-%m-%d %H:%M}"
        report + time


@click.argument("event", default=None, required=False)
@report.command()
def status(event):
    """
    Provide a simple summary of the status of a given event.

    Arguments
    ---------
    name : str, optional
       The name of the event.

    """
    for event in sorted(current_ledger.get_event(event), key=lambda e: e.name):
        click.secho(f"{event.name:30}", bold=True)
        if len(event.productions) > 0:
            click.secho("\tAnalyses", bold=True)
            if len(event.productions) == 0:
                click.echo("\t<NONE>")
            for production in event.productions:
                click.echo(
                    f"\t- {production.name} "
                    + click.style(f"{production.pipeline}")
                    + " "
                    + click.style(f"{production.status}")
                )
        if len(event.get_all_latest()) > 0:
            click.secho(
                "\tAnalyses waiting: ",
                bold=True,
            )
            waiting = event.get_all_latest()
            for awaiting in waiting:
                click.echo(
                    f"{awaiting.name} ",
                )
            click.echo("")


@click.option(
    "--yaml", "yaml_f", default=None, help="A YAML file to save the ledger to."
)
@click.argument("event", default=None, required=False)
@report.command()
def ledger(event, yaml_f):
    """
    Return the ledger for a given event.
    If no event is specified then the entire production ledger is returned.
    """
    total = []
    for event in current_ledger.get_event(event):
        total.append(yaml.safe_load(event.to_yaml()))

    click.echo(yaml.dump(total))

    if yaml_f:
        with open(yaml_f, "w") as f:
            f.write(yaml.dump(total))



"""
Review functions for asimov events.
"""

import os

import click

from asimov import config, current_ledger
from asimov.pipelines import known_pipelines
from asimov.review import ReviewMessage


@click.group()
def review():
    """Add and view review information and sign-offs"""
    pass


@click.option("--message", "-m", "message", default=None)
@click.argument("status", required=False, default=None)
@click.argument("production", required=True)
@click.argument("event", required=True)
@review.command()
def add(event, production, status, message):
    """
    Add a review signoff or rejection to an event.
    """
    valid = {"REJECTED", "APPROVED", "PREFERRED", "DEPRECATED"}
    events = current_ledger.get_event(event)
    if events is None:
        click.echo(
            click.style("●", fg="red") + f" Could not find an event called {event}"
        )
    else:
        for event in events:
            production = [
                production_o
                for production_o in event.productions
                if production_o.name == production
            ][0]
            click.secho(event.name, bold=True)

            if status.upper() in valid:
                message = ReviewMessage(
                    message=message, status=status, production=production
                )
                production.review.add(message)
            else:
                click.echo(
                    click.style("●", fg="red")
                    + f" Did not understand the review status {status.lower()}."
                    + " The review status must be one of "
                    + "{APPROVED, REJECTED, PREFERRED, DEPRECATED}"
                )

            if hasattr(event, "issue_object"):
                production.event.update_data()
            current_ledger.update_event(event)
            click.echo(
                click.style("●", fg="green")
                + f" {event.name}/{production.name} {status.lower()}"
            )


@click.argument("production", default=None, required=False)
@click.argument("event", default=None, required=False)
@review.command()
def status(event, production):
    """
    Show the review status of an event.
    """
    for event in current_ledger.get_event(event):
        click.secho(event.name, bold=True)
        if production:
            productions = [
                prod for prod in event.productions if prod.name == production
            ]
        else:
            productions = event.productions

        for production in productions:
            click.secho(f"\t{production.name}", bold=True)
            if production.review:
                click.echo(f"\t\t {production.review.status.lower()}")
            else:
                click.secho("\t\tNo review information exists for this production.")


@click.argument("event", default=None, required=False)
@review.command()
def audit(event):
    """
    Conduct an audit of the contents of production ini files
    against the production ledger.

    Parameters
    ----------
    event : str, optional
       The event to be checked.
       Optional; if the event isn't provided all events will be audited.
    """
    if isinstance(event, str):
        event = [event]

    for production in current_ledger.get_event(event)[0].productions:
        category = config.get("general", "calibration_directory")
        config_file = os.path.join(
            production.event.repository.directory, category, f"{production.name}.ini"
        )
        pipe = known_pipelines[production.pipeline.lower()](production, category)
        click.echo(pipe.read_ini(config_file))



"""
Make the event issues on the gitlab issue tracker from gracedb.
"""
import os

import asimov
from asimov.event import Event, DescriptionException
from asimov import config

from asimov import gitlab
from asimov import config

import numpy as np
import yaml


import click


server = gitlab.gitlab.Gitlab(config.get("gitlab", "url"), private_token=config.get("gitlab", "token"))
repository = server.projects.get(config.get("olivaw", "tracking_repository"))



def find_calibrations(time):
    with open("LLO_calibs.txt") as llo_file:
        data_llo = llo_file.read().split("\n")
        data_llo = [datum for datum in data_llo if datum[-16:]=="FinalResults.txt"]
        times_llo = {int(datum.split("GPSTime_")[1].split("_C01")[0]): datum for datum in data_llo}
    
    with open("LHO_calibs.txt") as llo_file:
        data_lho = llo_file.read().split("\n")
        data_lho = [datum for datum in data_lho if datum[-16:]=="FinalResults.txt"]
        times_lho = {int(datum.split("GPSTime_")[1].split("_C01")[0]): datum for datum in data_lho}
        
    keys_llo = np.array(list(times_llo.keys())) 
    keys_lho = np.array(list(times_lho.keys())) 
    return {"H1": times_lho[keys_lho[np.argmin(np.abs(keys_lho - time))]], "L1": times_llo[keys_llo[np.argmin(np.abs(keys_llo - time))]]}


@click.group()
def olivaw():
    """
    This is the main olivaw program which runs the DAGs for each event issue.
    """
    click.echo("Running olivaw")

    global rundir
    rundir = os.getcwd()

@click.option("--event", "event", default=None, help="The event which the ledger should be returned for, optional.")
@olivaw.command()
def notes(event):
    gitlab_events = gitlab.find_events(repository, subset=event)
    # Update existing events
    delete_list = []
    for event in gitlab_events:
        for note in event.issue_object.notes.list():
            if "error was detected" in note.body:
                delete_list.append(note)
        for note in delete_list:
            note.delete()

olivaw()



import ast
import subprocess, os
from shutil import copy
from asimov import gitlab, mattermost
from asimov import config
from asimov import condor, git
from asimov.ini import RunConfiguration

server = gitlab.gitlab.Gitlab('https://git.ligo.org', private_token=config.get("gitlab", "token"))
repository = server.projects.get(config.get("olivaw", "tracking_repository"))


uber_repository = git.MetaRepository(config.get("olivaw", "metarepository"))

events = gitlab.find_events(repository, milestone=config.get("olivaw", "milestone"))

from pesummary.gw.file.read import read


class SummaryTable(object):
    """
    Construct a summary table which describes the contents of a PE Summary file.
    """

    def __init__(self, event, summaryfile):
        """
        Load a PE Summary file to... summarise..

        Parameters
        ----------
        event : EventIssue
           The event issue for this event.
        summaryfile : str or path-like object
           The location of the summary file to be summarised.
        """

        self.summaryfile = read(summaryfile)
        self.event = event
        self.webdir = f"https://ldas-jobs.ligo.caltech.edu/~daniel.williams/LVC/projects/O3/preferred/{event.title}"

    def summarise(self):
        """
        Produce a markdown-formatted output which fully describes this summary.
        """
        
        output = """"""
        output += f"# {self.event.title}\n"

        output += """
| Run  | Code Version | Approx | IFOs | flow (Hz) | Results | Samples | PSDs | Calibration | Notes | Contact 
|----- | -----------  | ------ | ---- | --------  | -----  | -----  | ----- | -----  | -----  | -----|\n"""
        priors = {}
        for label in self.summaryfile.labels:
            run_config = RunConfiguration(self.summaryfile.config[label])
            engine = run_config.get_engine()
            priors[label] = engine
            flow = run_config.ini.get("lalinference", "flow")
            psds = len(self.summaryfile.psd[label])
            calibration = len(self.summaryfile.calibration[label])
            output += f"| {label} | | {engine['approx']} | {run_config.get_ifos()} | {flow} | | | {psds} | {calibration} | | daniel.williams | \n"

        output += "\n## Priors\n"
        for label in self.summaryfile.labels:
            output += f"## {label}\n"
            output += f"```\n"
            for key, value in priors[label].items():
                output += f"{key} = {value}\n"
            output += "```\n"


        return output
        

    def __repr__(self):
        return self.summarise()
    
    def __str__(self):
        return self.summarise()

def main():
    
    for event in events:
        print(event.title)
        if not "Preferred cleaned" in event.labels:
           continue

        try:
            repo = uber_repository.get_repo(event.title)
        except:
            print(f"{event.title} missing from the uberrepo")
            continue

        try:
            preferred_file = os.path.join(repo.directory, "Preferred", "PESummary_metafile", "posterior_samples.h5")
            preferred_readme = os.path.join(repo.directory, "Preferred", "README.md")
            table = SummaryTable(event=event, summaryfile=preferred_file)
        except OSError as e:
            print(e)
            continue

        readme_path = os.path.join(repo.directory, "Preferred", "README.md")


        with open(readme_path, "w") as preferred_readme:
                  print(preferred_readme.write(table.summarise()))

        try:
            repo.repo.git.add(readme_path)
            repo.repo.git.commit(m=":robot: Updated the preferred directory readme.")
            repo.repo.git.push()
        except git.git.exc.GitCommandError as e:
            print(e)

main()



"""
Script to add event data from an ini configuration file.
"""
import ast

import asimov

from asimov.event import Event, DescriptionException
from asimov import config
from asimov import gitlab



import numpy as np
from configparser import ConfigParser, NoOptionError

from git.exc import GitCommandError
import click
import collections.abc


server = gitlab.gitlab.Gitlab(config.get("gitlab", "url"), private_token=config.get("gitlab", "token"))
repository = server.projects.get(config.get("olivaw", "tracking_repository"))


@click.option("--event", "event", default=None, help="The event which the ledger should be returned for, optional.")
@click.option("--ini", "ini_data", default=None)
@click.command()
def add_data(event, ini_data=None):    
    gitlab_event = gitlab.find_events(repository, subset=event)
    def update(d, u):
        for k, v in u.items():
            if isinstance(v, collections.abc.Mapping):
                d[k] = update(d.get(k, {}), v)
            else:
                d[k] = v
        return d

    ini = ConfigParser()
    ini.optionxform=str

    ini.read(ini_data)

    new_data = {"quality": {}, "priors": {}}
    new_data["quality"]["sample-rate"] = int(ini.get("engine", "srate"))
    new_data["quality"]["lower-frequency"] = ast.literal_eval(ini.get("lalinference", "flow"))
    new_data["quality"]["segment-length"] = float(ini.get("engine", "seglen"))
    new_data["quality"]["window-length"] = float(ini.get("engine", "seglen"))
    new_data["quality"]["psd-length"] = float(ini.get("engine", "seglen"))
    new_data["quality"]["reference-frequency"] = float(ini.get("engine", "fref"))

    # new_data["priors"]["amp order"] = data['amp_order']
    try: 
        ini.get("engine", "comp-max")
        new_data["priors"]["component"] = [float(ini.get("engine", "comp-min")), 
                                           float(ini.get("engine", "comp-max"))]
    except:
        pass
        #new_data["priors"]["component"] = [float(ini.get("engine", "comp-min")), 
        #                                   None]

    try:
        new_data["priors"]["chirp-mass"] = [float(ini.get("engine", "chirpmass-min")), 
                                            float(ini.get("engine", "chirpmass-max"))]
        new_data["priors"]["chirp-mass"] = [None, ini.get("engine", "distance-max")]
    except:
        pass
    new_data["priors"]["q"] = [float(ini.get("engine", "q-min")), 1.0]


    update(gitlab_event[0].event_object.meta, new_data)
    print(gitlab_event[0].event_object.meta)
    gitlab_event[0].update_data()

add_data()



import collections
import glob
import os
from contextlib import contextmanager
from copy import deepcopy
from pathlib import Path

import numpy as np

from asimov import logger


@contextmanager
def set_directory(path: (Path, str)):
    """
    Change to a different directory for the duration of the context.

    Args:
        path (Path): The path to the cwd

    Yields:
        None
    """

    origin = Path().absolute()
    try:
        logger.info(f"Working temporarily in {path}")
        os.chdir(path)
        yield
    finally:
        os.chdir(origin)
        logger.info(f"Now working in {origin} again")


def find_calibrations(time):
    """
    Find the calibration file for a given time.
    """
    if time < 1190000000:
        dir = "/home/cal/public_html/uncertainty/O2C02"
        virgo = "/home/carl-johan.haster/projects/O2/C02_reruns/V_calibrationUncertaintyEnvelope_magnitude5p1percent_phase40mraddeg20microsecond.txt"  # NoQA
    elif time < 1290000000:
        dir = "/home/cal/public_html/uncertainty/O3C01"
        virgo = "/home/cbc/pe/O3/calibrationenvelopes/Virgo/V_O3a_calibrationUncertaintyEnvelope_magnitude5percent_phase35milliradians10microseconds.txt"  # NoQA
    elif time > 1268306607:
        raise ValueError(
            "This method cannot be used to add calibration envelope data beyond O3."
        )
    data_llo = glob.glob(f"{dir}/L1/*LLO*FinalResults.txt")
    times_llo = {
        int(datum.split("GPSTime_")[1].split("_C0")[0]): datum for datum in data_llo
    }

    data_lho = glob.glob(f"{dir}/H1/*LHO*FinalResults.txt")
    times_lho = {
        int(datum.split("GPSTime_")[1].split("_C0")[0]): datum for datum in data_lho
    }

    keys_llo = np.array(list(times_llo.keys()))
    keys_lho = np.array(list(times_lho.keys()))

    return {
        "H1": times_lho[keys_lho[np.argmin(np.abs(keys_lho - time))]],
        "L1": times_llo[keys_llo[np.argmin(np.abs(keys_llo - time))]],
        "V1": virgo,
    }


def update(d, u, inplace=True):
    """Recursively update a dictionary."""
    if not inplace:
        d = deepcopy(d)
        u = deepcopy(u)
    for k, v in u.items():
        if isinstance(v, collections.abc.Mapping):
            d[k] = update(d.get(k, {}), v)
        else:
            d[k] = v
    return d


# The following function adapted from https://stackoverflow.com/a/69908295
def diff_dict(d1, d2):
    d1_keys = set(d1.keys())
    d2_keys = set(d2.keys())
    shared_keys = d1_keys.intersection(d2_keys)
    shared_deltas = {o: (d1[o], d2[o]) for o in shared_keys if d1[o] != d2[o]}
    added_keys = d2_keys - d1_keys
    added_deltas = {o: (None, d2[o]) for o in added_keys}
    deltas = {**shared_deltas, **added_deltas}
    return parse_deltas(deltas)


# The following function adapted from https://stackoverflow.com/a/69908295
def parse_deltas(deltas: dict):
    res = {}
    for k, v in deltas.items():
        if isinstance(v[0], dict):
            tmp = diff_dict(v[0], v[1])
            if tmp:
                res[k] = tmp
        else:
            res[k] = v[1]
    return res



"""
The locutus script.
"""
import os

import click

from asimov.storage import AlreadyPresentException, NotAStoreError, Store

cwd = os.getcwd()
try:
    this_store = Store(root=cwd)
except NotAStoreError:
    this_store = None


@click.group()
def cli():
    pass


@click.option(
    "--name", "name", prompt="Name for the store: ", help="The name for the store."
)
@cli.command()
def init(name):
    """
    Initialise a new Store
    """

    # Get the working directory
    cwd = os.getcwd()
    Store.create(cwd, name)


@cli.command()
def info():
    """
    Check the information about this Store.
    """
    cwd = os.getcwd()
    store = Store(root=cwd)

    click.echo(store.manifest.data)


@click.argument("filename")  # , help="The file to add.")
@click.argument("production")  # , help="The production name.")
@click.argument("event")  # , help="The event label.")
@cli.command()
def store(event, production, filename):
    """
    Store a file in the Store.
    """
    try:
        click.echo(this_store.add_file(event, production, filename))
    except AlreadyPresentException:
        click.echo("This resource has already been stored.")


@click.option(
    "--hash", "hash", help="Optional hash to check the returned file against."
)
@click.argument("filename")
@click.argument("production")
@click.argument("event")
@cli.command()
def fetch(event, production, filename, hash=None):
    """
    Fetch a file from an event production.
    """
    try:
        click.echo(this_store.fetch_file(event, production, filename, hash))
    except FileNotFoundError:
        click.echo("File not found in the manifest.")


@click.argument("production")
@click.argument("event")
@cli.command()
def list(event, production):
    """
    Store a file in the Store.
    """
    click.echo(f"{'Resource':30} {'Hash':32} {'UUID':32}")
    click.echo("-" * 96)
    for resource, details in this_store.manifest.list_resources(
        event, production
    ).items():
        click.echo(f"{resource:30} {details['hash']:32} {details['uuid']:32}")



"""These tests are designed to be run on all pipelines to ensure that
they are generally compliant within asimov's specifications.  They can
include checking that ini files have the correct information and that
the pipelines report information as expected.
"""

import os
import io
import unittest
import shutil
import git
import contextlib

import asimov.event
from asimov.cli.project import make_project
from asimov.cli.application import apply_page
from asimov.ledger import YAMLLedger
from asimov.pipelines import known_pipelines

from asimov.testing import AsimovTestCase
from asimov import config
from asimov.utils import set_directory

exemplar = {
    "bayeswave": {
        "minimum frequency": "62",
        "segment length": "4",
        "psd length": "108.6",
        "ifo list": "['H1', 'L1', 'V1']",
        "segment start": "2",
        "data channels": "{ 'H1':'H1:WeirdChannel','L1':'L1:WeirdChannel','V1':'V1:OddChannel', }",
        "data frames": "{ 'H1':'NonstandardFrame','L1':'NonstandardFrameL1','V1':'UnusualFrameType', }",
        "window length": "71",
    },
    "bilby": {
        "minimum frequency": "{ H1:62,L1:92,V1:62, }",
        "segment length": "4",
        "psd length": "109.0", # Note that bilby is set-up to round the PSD length to the nearest second
        "ifo list": "['H1', 'L1', 'V1']",
        "segment start": "2",
        "data channels": "{ H1:WeirdChannel,L1:WeirdChannel,V1:OddChannel, }",
        "data frames": "{ 'H1':'NonstandardFrame','L1':'NonstandardFrameL1','V1':'UnusualFrameType', }",
        "window length": "71",
    },
    "rift": {
        "minimum frequency": '{ "H1":62,"L1":92,"V1":62, }',
        "segment length": "4",
        "psd length": "108.6",
        "ifo list": "['H1', 'L1', 'V1']",
        "data channels": '{ "H1":"H1:WeirdChannel","L1":"L1:WeirdChannel","V1":"V1:OddChannel", }',
        "data frames": '{ "H1":"NonstandardFrame","L1":"NonstandardFrameL1","V1":"UnusualFrameType", }',
        "window length": "71",
    },
    "lalinference": {
        "minimum frequency": "{ 'H1': 62, 'L1': 92,  'V1': 62  }",
        "segment length": "4",
        "ifo list": "['H1', 'L1', 'V1']",
        "data channels": "{'H1': 'H1:WeirdChannel', 'L1': 'L1:WeirdChannel', 'V1': 'V1:OddChannel'}",
        "data frames": "{'H1': 'NonstandardFrame', 'L1': 'NonstandardFrameL1', 'V1': 'UnusualFrameType'}",
    }
}


mappings = {
    "bayeswave": {
        "minimum frequency": ["input", "flow"],
        "segment length": ["input", "seglen"],
        "window length": ["input", "window"],
        "psd length": ["input", "psdlength"],
        "ifo list": ["input", "ifo-list"],
        "segment start": ["input", "segment-start"],
        "data channels": ["datafind", "channel-list"],
        "data frames": ["datafind", "frtype-list"],
    },
    "bilby": {
        "minimum frequency": ["test", "minimum-frequency"],
        "segment length": ["test", "duration"],
        "psd length": ["test", "psd-length"],
        "ifo list": ["test", "detectors"],
        "data channels": ["test", "channel-dict"],
    },
    "rift": {
        "data channels": ["datafind", "channel-list"],
        "data frames": ["datafind", "frtype-list"],
        "minimum frequency": ["lalinference", "flow"],
        "segment length": ["engine", "seglen"],
        "ifo list": ["analysis", "ifos"],
        "data channels": ["data", "channels"],
        "data frames": ["datafind", "types"],
    },
    "lalinference": {
        "minimum frequency": ["lalinference", "flow"],
        "segment length": ["engine", "seglen"],
        "ifo list": ["analysis", "ifos"],
        "data channels": ["data", "channels"],
        "data frames": ["datafind", "types"],
    },
}


class TestIniFileHandling(AsimovTestCase):
    """Test that for all pipelines the information in the ini file is
    correctly substituted."""

    def test_detchar_substitution(self):
        event = "Nonstandard fmin"

        for pipeline in known_pipelines.keys():
            if pipeline not in exemplar.keys():
                continue
            if pipeline == "bilby": continue
            f = io.StringIO()
            with contextlib.redirect_stdout(f):

                apply_page(
                    f"{self.cwd}/tests/test_data/testing_pe.yaml",
                    event=None,
                    ledger=self.ledger,
                )
                apply_page(
                    f"{self.cwd}/tests/test_data/event_non_standard_settings.yaml",
                    event=None,
                    ledger=self.ledger,
                )

            f = io.StringIO()
            with contextlib.redirect_stdout(f):
                apply_page(
                    f"{self.cwd}/tests/test_data/{pipeline}_settings_nodeps.yaml",
                    event=event,
                    ledger=self.ledger,
                )

            f = io.StringIO()
            with contextlib.redirect_stdout(f):
                production = self.ledger.get_event(event)[0].productions[0]
                with set_directory(
                    os.path.join(
                        "checkouts", event, config.get("general", "calibration_directory")
                    )
                ):
                    production.make_config(f"{production.name}.ini")
            ini_location = os.path.join(
                config.get("project", "root"),
                "checkouts",
                event,
                config.get("general", "calibration_directory"),
                f"{production.name}.ini",
            )
            self.assertTrue(os.path.exists(ini_location))

            # Read the ini file and convert it to a dictionary for easy access

            import configparser

            pipeline_config = configparser.ConfigParser()
            try:
                pipeline_config.read(ini_location)
            except configparser.MissingSectionHeaderError:
                with open(ini_location, "r") as file_handle:
                    data = file_handle.read()
                with open(ini_location, "w") as file_handle:
                    data = """[test]
""" + data
                    file_handle.write(data)
                pipeline_config.read(ini_location)

            pipeline_config = {
                s: dict(pipeline_config.items(s)) for s in pipeline_config.sections()
            }

            for name, mapping in mappings[pipeline].items():
                with self.subTest(name=name, pipeline=pipeline):
                    self.assertEqual(
                        pipeline_config[mapping[0]][mapping[1]], exemplar[pipeline][name]
                    )



import ast
import json
import os
from math import floor

import click
import gwpy
import gwpy.timeseries
import numpy as np
from glue.lal import Cache
from gwdatafind import find_urls

from asimov import config
from asimov import current_ledger as ledger
from asimov.utils import find_calibrations, update
from asimov.event import DescriptionException, Event


@click.group()
def event():
    """
    Commands to handle events & collections.
    """
    pass


@click.option("--search", "-s", "search", default=None, help="Search criteria.")
@click.option(
    "--old", "oldname", default=None, help="The old superevent ID for this event."
)
@click.option(
    "--gid",
    "gid",
    default=None,
    help="The GraceDB GID for the event (for legacy events)",
)
@click.option(
    "--superevent", "superevent", default=None, help="The superevent for the event."
)
@click.option(
    "--repository",
    "repo",
    default=None,
    help="The location of the repository for this event.",
)
@click.option("--name", "-n", "name", default=None, help="The name for the event.")
@event.command()
def create(name=None, oldname=None, gid=None, superevent=None, repo=None, search=None):
    """
    Create a new event record in the ledger.

    Parameters
    ----------
    superevent : str
       The ID of the superevent to be used from GraceDB
    name : str
       The name of the event to be recorded in the issue tracker
    names : path, optional
        The path to the name file which maps between old and new super event IDs
    oldname : str, optional
        The old name of the event.
    search : str, optional
        A string of search criteria
    """
    import pathlib

    if gid or superevent or search:
        from ligo.gracedb.rest import GraceDb

        client = GraceDb(service_url=config.get("gracedb", "url"))

    if search:
        event_iterator = client.events(search)

        for event in event_iterator:

            event = Event(
                name=name,
                repository=repo,
                calibration={},
            )
            working_dir = os.path.join(config.get("general", "rundir_default"), name)

            event.meta["working directory"] = working_dir
            pathlib.Path(working_dir).mkdir(parents=True, exist_ok=True)
            ledger.update_event(event)

    if superevent:
        data = client.superevent(superevent).json()
        try:
            event_data = client.event(data["preferred_event"]).json()
        except KeyError:
            event_data = data["preferred_event_data"]

        gid = data["preferred_event_data"]["graceid"]
        interferometers = event_data["instruments"].split(",")
        if not name:
            name = superevent
    elif gid:
        event_data = client.event(gid).json()
        interferometers = event_data["instruments"].split(",")
        if not name:
            name = gid
    else:
        event_data = None
        interferometers = []

    if not repo:
        repo = None

    event = Event(
        name=name,
        repository=repo,
        calibration={},
        interferometers=interferometers,
    )

    if oldname:
        event.meta["old superevent"] = oldname
    if gid:
        event.meta["event time"] = event_data["gpstime"]

    if "ligo" not in event.meta:
        event.meta["ligo"] = {}

    event.meta["ligo"]["preferred event"] = gid

    working_dir = os.path.join(config.get("general", "rundir_default"), name)

    event.meta["working directory"] = working_dir
    pathlib.Path(working_dir).mkdir(parents=True, exist_ok=True)
    ledger.update_event(event)


@click.argument("delete")
@event.command()
def delete(event):
    """
    Delete an event from the ledger.
    """
    ledger.delete_event(event_name=event)


# @click.argument("event")
# @click.option("--yaml", "yaml", default=None)
# @click.option("--ini", "ini", default=None)
# @event.command()
# def populate(event, yaml, ini):
#     """
#     Populate an event ledger with data from ini or yaml files.
#     """

#     event = ledger.get_event(event)
#     # Check the calibration files for this event
#     click.echo("Check the calibration.")
#     click.echo(event.name)
#     calibration(event=event.name)
#     # Check the IFOs for this event
#     click.echo("Check the IFO list")
#     try:
#         checkifo(event.name)
#     except:
#         pass

#     if yaml:
#         add_data(event.name, yaml)


@click.argument("event", default=None)
@click.option("--json", "json_data", default=None)
@event.command()
def configurator(event, json_data=None):
    """Add data from the configurator."""
    event = ledger.get_event(event)
    if json_data:
        with open(json_data, "r") as datafile:
            data = json.load(datafile)

    new_data = {"quality": {}, "priors": {}}
    new_data["quality"]["sample-rate"] = int(data["srate"])
    new_data["quality"]["lower-frequency"] = {}
    # Factor 0.875 to account for PSD roll off
    new_data["likelihood"]["upper-frequency"] = {
        ifo: int(0.875 * data["srate"] / 2) for ifo in event.meta["interferometers"]
    }
    new_data["quality"]["start-frequency"] = data["f_start"]
    new_data["quality"]["segment-length"] = int(data["seglen"])
    new_data["quality"]["window-length"] = int(data["seglen"])
    new_data["quality"]["psd-length"] = int(data["seglen"])

    def decide_fref(freq):
        if (freq >= 5) and (freq < 10):
            return 5
        else:
            return floor(freq / 10) * 10

    new_data["quality"]["reference-frequency"] = decide_fref(data["f_ref"])

    new_data["priors"]["amp order"] = data["amp_order"]
    new_data["priors"]["chirp-mass"] = [data["chirpmass_min"], data["chirpmass_max"]]

    update(event.meta, new_data)
    ledger.update_event(event)


# @click.option("--event", "event", default=None, help="The event which the ledger should be returned for, optional.")
# @olivaw.command()
@click.argument("event")
@event.command()
def checkifo(event):
    event = ledger.get_event(event)
    if "event time" not in event.event_object.meta:
        print(f"Time not found {event.event_object.name}")
    time = event.event_object.meta["event time"]
    gpsstart = time - 600
    gpsend = time + 600
    bits = ["Bit 0", "Bit 1", "Bit 2"]

    active_ifo = []
    for ifo in ["L1", "H1", "V1"]:
        frametypes = event.event_object.meta["data"]["frame-types"]
        urls = find_urls(
            site=f"{ifo[0]}",
            frametype=frametypes[ifo],
            gpsstart=gpsstart,
            gpsend=gpsend,
        )
        datacache = Cache.from_urls(urls)
        if len(datacache) == 0:
            print(f"No {ifo} data found.")
            continue

        if "state vector" in event.meta:
            state_vector_channel = event.meta["state vector"]
        else:
            state_vector_channel = ast.literal_eval(config.get("data", "state-vector"))

        state = gwpy.timeseries.StateVector.read(
            datacache,
            state_vector_channel[ifo],
            start=gpsstart,
            end=gpsend,
            pad=0,  # padding data so that errors are not raised even if found data are not continuous.
        )
        if not np.issubdtype(state.dtype, np.unsignedinteger):
            # if data are not unsigned integers, cast to them now so that
            # we can determine the bit content for the flags
            state = state.astype(
                "uint32",
                casting="unsafe",
                subok=True,
                copy=False,
            )
        flags = state.to_dqflags()

        segments = flags[bits[0]].active
        for bit in bits:
            segments -= ~flags[bit].active

        if len(segments) > 0:
            active_ifo += [ifo]
    click.echo(event.name)
    if event.meta["interferometers"] != active_ifo:
        print(f"Gitlab data\t{event.meta['interferometers']}")
        print(f"Recommended IFOS\t{active_ifo}")

    event.meta["interferometers"] = active_ifo
    ledger.update_event(event)


@click.option(
    "--calibration",
    "calibration",
    multiple=True,
    default=[None],
    help="The location of the calibration files.",
)
@click.argument("event")
@event.command()
def calibration(event, calibration):
    event = ledger.get_event(event)[0]
    try:
        event._check_calibration()
    except DescriptionException:
        print(event.name)
        time = event.meta["event time"]
        if not calibration[0]:
            try:
                calibrations = find_calibrations(time)
            except ValueError:
                calibrations = {}
        else:
            calibrations = {}
            for cal in calibration:
                calibrations[cal.split(":")[0]] = cal.split(":")[1]
        print(calibrations)
        update(event.meta["data"]["calibration"], calibrations)
        ledger.update_event(event)


# @click.argument("data")
# @click.argument("event")
# @event.command()
# def load(event, data):
#     event = ledger.get_event(event)

#     with open(data, "r") as datafile:
#         data = yaml.safe_load(datafile.read())

#         event.meta = update(event.meta, data)

#     ledger.update_event(event)



"""Tests for the RIFT interface."""

import unittest
import shutil
import os
import git

import io
import contextlib

from unittest.mock import patch
from importlib import reload

from click.testing import CliRunner

import asimov
from asimov.cli import project, manage
from asimov.cli import configuration
from asimov.cli.application import apply_page
from asimov.ledger import YAMLLedger
from asimov.pipelines.rift import Rift
from asimov.event import Event
from asimov.pipeline import PipelineException


# @unittest.skip("Skipped until RIFT is added to the testing environment correctly.")
class RiftTests(unittest.TestCase):
    """Test RIFT interface.

    The test_dag method will need to be updated.
"""

    @classmethod
    def setUpClass(cls):
        cls.cwd = os.getcwd()

    def setUp(self):
        os.makedirs(f"{self.cwd}/tests/tmp/project")
        os.chdir(f"{self.cwd}/tests/tmp/project")
        f = io.StringIO()
        with contextlib.redirect_stdout(f):
        
            runner = CliRunner()
            result = runner.invoke(project.init,
                                   ['Test Project', '--root', f"{self.cwd}/tests/tmp/project"])
            assert result.exit_code == 0
            assert result.output == '● New project created successfully!\n'
            self.ledger = YAMLLedger()

    def tearDown(self):
        os.chdir(self.cwd)
        shutil.rmtree(f"{self.cwd}/tests/tmp/project/")

    # @unittest.skip("Skipped temporarily while RIFT is updated")
    def test_submit_cli(self):
        """Check that a RIFT config file can be built."""
        f = io.StringIO()
        with contextlib.redirect_stdout(f):
            apply_page(file = f"https://git.ligo.org/asimov/data/-/raw/main/defaults/production-pe.yaml", event=None, ledger=self.ledger)
            apply_page(file = "https://git.ligo.org/asimov/data/-/raw/main/defaults/production-pe-priors.yaml", event=None, ledger=self.ledger)
            event = "GW150914_095045"
            pipeline = "rift"
            apply_page(file = f"https://git.ligo.org/asimov/data/-/raw/main/tests/{event}.yaml", event=None, ledger=self.ledger)
            apply_page(file = f"{self.cwd}/tests/test_data/test_{pipeline}.yaml", event=event, ledger=self.ledger)
        with patch("asimov.current_ledger", new=YAMLLedger()):
            reload(asimov)
            reload(manage)
            runner = CliRunner()
            result = runner.invoke(manage.manage, ['build'])
            result = runner.invoke(manage.submit, "--dryrun")
        self.assertTrue("util_RIFT_pseudo_pipe.py" in result.output)

    def test_build_api(self):
        """Check that a RIFT config file can be built."""
        f = io.StringIO()
        with contextlib.redirect_stdout(f):
            apply_page(file = "https://git.ligo.org/asimov/data/-/raw/main/defaults/production-pe.yaml", event=None, ledger=self.ledger)
            apply_page(file = "https://git.ligo.org/asimov/data/-/raw/main/defaults/production-pe-priors.yaml", event=None, ledger=self.ledger)
            event = "GW150914_095045"
            pipeline = "rift"
            apply_page(file = f"https://git.ligo.org/asimov/data/-/raw/main/tests/{event}.yaml", event=None, ledger=self.ledger)
            apply_page(file = f"{self.cwd}/tests/test_data/test_{pipeline}.yaml", event=event, ledger=self.ledger)
        with patch("asimov.current_ledger", new=YAMLLedger()):
            reload(asimov)
            reload(manage)
            runner = CliRunner()
            result = runner.invoke(manage.manage, ['build'])
            result = runner.invoke(manage.submit, "--dryrun")
        f = io.StringIO()
        with contextlib.redirect_stdout(f):
            result = self.ledger.get_event(event)[0].productions[0].pipeline.build_dag(dryrun=True)
            print("\n\nBUILDDAG ",f.getvalue())
            print("RESULT", result)
            self.assertTrue("util_RIFT_pseudo_pipe.py --assume-nospin --calibration C01 --approx IMRPhenomD" in f.getvalue())
            self.assertTrue("--ile-force-gpu " in f.getvalue())
        

    def test_build_api_non_default_calibration(self):
        """Check that a RIFT correctly picks up non C01 calibration."""

        runner = CliRunner()
        result = runner.invoke(configuration.update,
                               ['--general/calibration_directory', f"C00_offline"])
        assert result.exit_code == 0
        result = runner.invoke(configuration.update,
                               ['--general/calibration', f"C00"])
        assert result.exit_code == 0

        f = io.StringIO()
        with contextlib.redirect_stdout(f):        
            apply_page(file = "https://git.ligo.org/asimov/data/-/raw/main/defaults/production-pe.yaml", event=None, ledger=self.ledger)
            apply_page(file = "https://git.ligo.org/asimov/data/-/raw/main/defaults/production-pe-priors.yaml", event=None, ledger=self.ledger)
            event = "GW150914_095045"
            pipeline = "rift"
            apply_page(file = f"https://git.ligo.org/asimov/data/-/raw/main/tests/{event}.yaml", event=None, ledger=self.ledger)
            apply_page(file = f"{self.cwd}/tests/test_data/test_{pipeline}.yaml", event=event, ledger=self.ledger)

        with patch("asimov.current_ledger", new=YAMLLedger()):
            reload(asimov)
            reload(manage)
            runner = CliRunner()
            result = runner.invoke(manage.manage, ['build'])
            result = runner.invoke(manage.submit, "--dryrun")
        f = io.StringIO()
        with contextlib.redirect_stdout(f):
            result = self.ledger.get_event(event)[0].productions[0].pipeline.build_dag(dryrun=True)
            print("\n\nBUILDDAG ",f.getvalue())
            print("RESULT", result)
            self.assertTrue("util_RIFT_pseudo_pipe.py --assume-nospin --calibration C00 --approx IMRPhenomD" in f.getvalue())
            self.assertTrue("--ile-force-gpu " in f.getvalue())



"""
Review-related code.

Note
----
This code does not directly relate to the review of *asimov*
but rather to the review of events which it has been used to analyse.
"""

from datetime import datetime

STATES = {"REJECTED", "APPROVED", "PREFERRED", "DEPRECATED"}
review_map = {
    "deprecated": "warning",
    "none": "default",
    "approved": "success",
    "rejected": "danger",
    "checked": "info",
}


class Review:
    """
    A class to record the review status of a given production.
    """

    def __init__(self):
        self.messages = []

    def __len__(self):
        return len(self.messages)

    def __getitem__(self, item):
        return self.messages[item]

    def add(self, review_message):
        """
        Add a new review message to an event.
        """
        self.messages.append(review_message)
        self.messages = sorted(self.messages, key=lambda k: k.timestamp)

    @property
    def status(self):
        status = None
        for message in self.messages:
            if message.status:
                status = message.status
        return status

    def to_dicts(self):
        output = []
        for message in self.messages:
            output.append(message.to_dict())
        return output

    @classmethod
    def from_dict(cls, messages_list, production):
        """parse a dictionary into review"""
        messages = []
        for message in messages_list:
            messages.append(
                ReviewMessage.from_dict(dictionary=message, production=production)
            )
        review_ob = cls()
        messages = sorted(messages, key=lambda k: k.timestamp)
        review_ob.messages = messages
        return review_ob


class ReviewMessage:
    """
    A review message.
    """

    def __init__(self, message, production, status=None, timestamp=None):
        """
        Review messages are individual messages related to the review of a production.

        Parameters
        ----------
        message : str
           The review message. This can be free-form text.
        production : `Asimov.event.Production`
           The production which this message is attached to.
        state: str, {"REJECTED", "APPROVED", "PREFERRED", "DEPRECATED"}
           The review state. This is optional, and a message can
           be added without a review state. If a review state is added
           it will over-ride any previously added review messages.
        timestamp: str, optional
           The timestamp for the message, optional, should be
           formatted "%Y-%m-%d %H:%M:%S".
        """

        self.message = message

        if status:
            if status.upper() in STATES:
                self.status = status.upper()
            else:
                raise ValueError(f"{status} is not a recognised review state.")
        else:
            self.status = None

        if timestamp:
            self.timestamp = (
                timestamp  # datetime.strptime(timestamp, "%Y-%m-%d %H:%M:%S")
            )
        else:
            self.timestamp = datetime.now()

    def to_dict(self):
        """
        Serialise this object as a dictionary
        """
        out = {}
        out["message"] = self.message
        out["timestamp"] = str(self.timestamp)
        out["status"] = self.status
        return out

    @classmethod
    def from_dict(cls, dictionary, production):
        """
        Create a review message from a dictionary.
        """
        default = {"status": None, "message": None, "timestamp": None}
        default.update(dictionary)
        if isinstance(default["timestamp"], str):
            timestamp = datetime.strptime(default["timestamp"], "%Y-%m-%d %H:%M:%S.%f")
        else:
            timestamp = default["timestamp"]
        message_ob = cls(
            message=default["message"],
            production=production,
            status=default["status"],
            timestamp=timestamp,
        )
        return message_ob

    def html(self):
        """
        Return an HTML representation of this review message.
        """
        review_row = ""
        if self.status:
            review_row += f"""<div class="asimov-review alert alert-{review_map[self.status.lower()]}" role="alert">
            <strong>{self.status}</strong>  """
        else:
            review_row += """<div class="alert alert-light" role="alert">"""
        if self.message:
            review_row += f"{self.message}"
        review_row += "</div>"
        return review_row



import unittest

import asimov
import asimov.condor

class CondorTests(unittest.TestCase):

# @classmethod
#     def setUpClass(cls):
#         cls.app = app = asimov.server.create_app()
#         app.config.update({
#         "TESTING": True,
#         })
#         cls.client = cls.app.test_client()


    def test_job_from_dict(self):
        """Check that a CondorJob object can be created from a dictionary."""

        dictionary = {"id": 450,
                      "command": "test.sh",
                      "hosts": "test.test.com",
                      "status": 1}

        job = asimov.condor.CondorJob.from_dict(dictionary)

        self.assertEqual(job.idno, 450)

    def test_status(self):
        """Check that status codes get translated to a human-readable string."""
        dictionary = {"id": 450,
                      "command": "test.sh",
                      "hosts": "test.test.com",
                      "status": 1}

        job = asimov.condor.CondorJob.from_dict(dictionary)

        self.assertEqual(job.status, "Idle")



"""
Project management tools.
"""

try:
    import ConfigParser as configparser
except ImportError:
    import configparser

import os
import shutil
import getpass

import click

from asimov import config, storage, logger, LOGGER_LEVEL
from asimov.ledger import Ledger

logger = logger.getChild("cli").getChild("project")
logger.setLevel(LOGGER_LEVEL)


def make_project(
    name,
    root,
    working="working",
    checkouts="checkouts",
    results="results",
    logs="logs",
    user=None,
):
    """
    Create a new project called NAME.

    This command creates a new asimov project, creating the appropriate
    directory structure, and creating a blank ledger.
    """
    import pathlib

    pathlib.Path(root).mkdir(parents=True, exist_ok=True)
    os.chdir(root)

    config.set("project", "name", name)
    config.set("project", "root", root)

    project_name = name

    # Make the virtual environment
    # builder = venv.EnvBuilder(system_site_packages=False,
    #                           clear=False,
    #                           symlinks=False,
    #                           upgrade=False,
    #                           with_pip=True,
    #                           prompt=f"Asimov {project_name}")

    # builder.create("environment")

    config.set("general", "environment", "environment")

    # Make the working directory
    pathlib.Path(working).mkdir(parents=True, exist_ok=True)
    config.set("general", "rundir_default", working)

    # Make the git directory
    pathlib.Path(checkouts).mkdir(parents=True, exist_ok=True)
    config.set("general", "git_default", checkouts)

    # Make the log directory
    pathlib.Path(logs).mkdir(parents=True, exist_ok=True)
    config.set("logging", "directory", logs)

    # Make the results store
    storage.Store.create(root=results, name=f"{project_name} storage")
    config.set("storage", "directory", results)

    # Make the ledger and operative files
    pathlib.Path(".asimov").mkdir(parents=True, exist_ok=True)
    config.set("ledger", "engine", "yamlfile")
    config.set("ledger", "location", os.path.join(".asimov", "ledger.yml"))

    # Set the default environment
    python_loc = shutil.which("python").split("/")[:-2]
    config.set("pipelines", "environment", os.path.join("/", *python_loc))
    config.set("rift", "environment", os.path.join("/", *python_loc))

    try:
        config.set("pesummary", "executable", shutil.which("summarypages"))
    except Exception:
        pass

    # Set the default condor user
    if not user:
        config.set("condor", "user", getpass.getuser())
    else:
        config.set("condor", "user", user)

    Ledger.create(
        engine="yamlfile",
        name=project_name,
        location=os.path.join(".asimov", "ledger.yml"),
    )

    with open(os.path.join(".asimov", "asimov.conf"), "w") as config_file:
        config.write(config_file)


@click.command()
@click.argument("name")
@click.option(
    "--root",
    default=os.getcwd(),
    help="Location to create the project, default is the current directory.",
)
@click.option(
    "--working",
    default="working",
    help="""The location to store working directories,
 default is a directory called 'working' inside the current directory.""",
)
@click.option(
    "--checkouts",
    default="checkouts",
    help="The location to store cloned git repositories.",
)
@click.option(
    "--results",
    default="results",
    help="The location where the results store should be created.",
)
@click.option(
    "--user",
    default=None,
    help="The user account to be used for accounting purposes. Defaults to the current user if not set.",
)
def init(
    name, root, working="working", checkouts="checkouts", results="results", user=None
):
    """
    Roll-out a new project.
    """
    make_project(name, root, working=working, checkouts=checkouts, results=results)
    click.echo(click.style("●", fg="green") + " New project created successfully!")
    logger.info(f"A new project was created in {os.getcwd()}")


@click.command()
@click.argument("location")
def clone(location):
    import pathlib
    import shutil

    results = "results"

    remote_config = os.path.join(location, "asimov.conf")
    config = configparser.ConfigParser()
    config.read([remote_config])
    click.echo(f'Cloning {config.get("project", "name")}')
    root = os.path.join(
        os.getcwd(), config.get("project", "name").lower().replace(" ", "-")
    )
    pathlib.Path(root).mkdir(parents=True, exist_ok=True)
    # os.chdir(root)
    config.set("project", "root", root)
    # Make the virtual environment
    # builder = venv.EnvBuilder(system_site_packages=False,
    #                          clear=False,
    #                          symlinks=False,
    #                          upgrade=False,
    #                          with_pip=True,
    #                          prompt=f"Asimov {project_name}")

    # builder.create("environment")

    # config.set("general", "environment", "environment")

    # Make the working directory
    # shutil.copytree(os.path.join(config.get("general", "rundir_default"), working)
    # config.set("general", "rundir_default", working)

    # Make the git directory
    # pathlib.Path(checkouts).mkdir(parents=True, exist_ok=True)
    # config.set("general", "git_default", checkouts)

    # Copy the results store
    # shutil.copyfile(os.path.join(location, config.get("storage", "results_store")), results)
    shutil.copytree(
        os.path.join(location, config.get("storage", "results_store")), results
    )
    config.set("storage", "results_store", results)

    # Make the ledger
    if config.get("ledger", "engine") == "yamlfile":
        shutil.copyfile(
            os.path.join(location, config.get("ledger", "location")),
            os.path.join(".asimov", "ledger.yml"),
        )
    elif config.get("ledger", "engine") == "gitlab":
        raise NotImplementedError(
            "The gitlab interface has been removed from this version of asimov."
        )

    config.set("ledger", "engine", "yamlfile")
    config.set("ledger", "location", os.path.join(".asimov", "ledger.yml"))

    with open(os.path.join(".asimov", "asimov.conf"), "w") as config_file:
        config.write(config_file)



from copy import copy

import click
import yaml

from asimov import config
from asimov import current_ledger as ledger

from asimov.event import Production
from asimov.storage import Store


@click.group()
def production():
    """
    Commands to handle productions.
    """
    pass


@click.argument("pipeline")
@click.argument("event")
@click.option(
    "--family",
    "family",
    default="Prod",
    help="The family name of the production, e.g. `prod`.",
)
@click.option(
    "--comment", "comment", default=None, help="A comment to attach to the production"
)
@click.option(
    "--needs",
    "needs",
    multiple=True,
    default=[],
    help="A list of productions which are requirements",
)
@click.option(
    "--template",
    "template",
    default=None,
    help="The configuration template for the production.",
)
@click.option(
    "--status", "status", default="ready", help="The initial status of the production."
)
@click.option(
    "--approximant",
    "approximant",
    default=None,
    help="The waveform approximant to be used in the production.",
)
@production.command(help="Add a new production to an event")
def create(event, pipeline, family, comment, needs, template, status, approximant):
    """
    Add a new production to an event.

    """

    event = ledger.get_event(event)[0]
    #
    event_prods = event.productions
    names = [production.name for production in event_prods]
    family_entries = [int(name.split(family)[1]) for name in names if family in name]
    #
    if "bayeswave" in needs:
        bw_entries = [
            production.name
            for production in event_prods
            if ("bayeswave" in production.pipeline.lower())
            and (production.review.status not in {"REJECTED", "DEPRECATED"})
        ]
        needs = bw_entries
    #
    production = {
        "comment": comment,
        "pipeline": pipeline,
        "status": status,
        "waveform": {},
    }
    if needs:
        production["needs"] = needs
    if template:
        production["template"] = template
    if approximant:
        production["waveform"]["approximant"] = approximant
    if len(family_entries) > 0:
        number = max(family_entries) + 1
    else:
        number = 0
    production_dict = {f"{family}{number}": production}
    production = Production.from_dict(production_dict, event=event)
    #
    event.add_production(production)
    click.secho(f"Production added to {event.name}")
    click.echo(f"\t{production}")

    if config.get("ledger", "engine") == "gitlab":
        raise NotImplementedError(
            "The Gitlab interface has been removed from this version." ""
        )
    elif config.get("ledger", "engine") == "yamlfile":
        ledger.events[event.name] = event.to_dict()
        ledger.save()


@click.option(
    "--status",
    "-s",
    default=None,
    help="Change the run status. Options: ready | wait | stuck | stop | cancelled",
)
@click.argument("production")
@click.argument("event")
@production.command()
def set(event, production, status):
    """
    Set or update properties such as the status of a production.
    Note that some properties cannot be updated once the production is being run,
    and you should create a new production instead.
    """
    event = ledger.get_event(event=event)[0]
    production = [
        production_o
        for production_o in event.productions
        if production_o.name == production
    ][0]

    accepted_states = {"ready", "wait", "stuck", "stop", "cancelled"}
    if status:
        if status in accepted_states:
            production.status = status
            # ledger.save()
            click.echo(
                click.style("●", fg="green")
                + f" {production.name} status updated to {status}"
            )
        else:
            click.echo(
                click.style("●", fg="red")
                + f" Unable to change the state of {production.name} to {status}."
            )


@click.argument("production")
@click.argument("event")
@production.command()
def show(event, production):
    """
    Show the entire metadata for a given production.
    """
    event = ledger.get_event(event=event)[0]
    production = [
        production_o
        for production_o in event.productions
        if production_o.name == production
    ][0]

    meta = copy(production.meta)
    meta.pop("ledger")
    output = yaml.dump(meta, default_flow_style=False)
    click.echo_via_pager(output)


@click.option("--file", "file", default=None)
@click.argument("production")
@click.argument("event")
@production.command()
def results(event, production, file, hash=None):
    """
    Fetch or list the results of a production.
    """
    event = ledger.get_event(event)[0]
    production = [
        production_o
        for production_o in event.productions
        if production_o.name == production
    ][0]
    store = Store(root=config.get("storage", "directory"))

    if not file:
        try:
            items = store.manifest.list_resources(event.name, production.name).items()
            click.secho(f"{'Resource':30} {'Hash':32} {'UUID':32}")
            click.secho("-" * 96)
            for resource, details in items:
                click.secho(f"{resource:30} {details['hash']:32} {details['uuid']:32}")
        except KeyError:
            click.secho("There are no results for this production.")
    else:
        try:
            click.echo(store.fetch_file(event, production, file, hash))
        except FileNotFoundError:
            click.secho(f"{file} could not be found for this production.")



import json
from math import floor
import click


@click.option("--event", "event", default=None, help="The event which will be updated")
@click.option("--json", "json_data", default=None)
@click.command(help="Add data from the configurator.")
def configurator(event, json_data=None):
    """
    Add data from the PEConfigurator script to an event.
    """

    if json_data:
        with open(json_data, "r") as datafile:
            data = json.load(datafile)

    new_data = {"quality": {}, "priors": {}}
    new_data["quality"]["sample-rate"] = data["srate"]
    new_data["quality"]["lower-frequency"] = {}
    new_data["quality"]["upper-frequency"] = int(0.875 * data["srate"] / 2)
    new_data["quality"]["start-frequency"] = data["f_start"]
    new_data["quality"]["segment-length"] = data["seglen"]
    new_data["quality"]["window-length"] = data["seglen"]
    new_data["quality"]["psd-length"] = data["seglen"]

    def decide_fref(freq):
        if (freq >= 5) and (freq < 10):
            return 5
        else:
            return floor(freq / 10) * 10

    new_data["quality"]["reference-frequency"] = decide_fref(data["f_ref"])

    new_data["priors"]["amp order"] = data["amp_order"]
    new_data["priors"]["chirp-mass"] = [data["chirpmass_min"], data["chirpmass_max"]]



"""
Make the event issues on the gitlab issue tracker from gracedb.
"""


import asimov
from asimov.event import Event
from asimov import config

from asimov import gitlab
from asimov import config

from ligo.gracedb.rest import GraceDb, HTTPError 
client = GraceDb(service_url=config.get("gracedb", "url"))
r = client.ping()

superevent_iterator = client.superevents('O3B_CBC_CATALOG')
superevent_ids = [superevent['superevent_id'] for superevent in superevent_iterator]

server = gitlab.gitlab.Gitlab(config.get("gitlab", "url"), private_token=config.get("gitlab", "token"))
repository = server.projects.get(config.get("olivaw", "tracking_repository"))


gitlab_events = gitlab.find_events(repository)

super_events = set(superevent_ids) - {event.title for event in gitlab_events}

# Add the new events
for superevent in list(super_events):

    data = client.superevent(superevent).json()
    event_data = client.event(data['preferred_event']).json()
    
    event_url = f"https://catalog-dev.ligo.org/events/{data['preferred_event']}/view/"
    
    event = Event(name=superevent,
                  repository=f"git@git.ligo.org:pe/O3/{superevent}",
                  gid=data['preferred_event'],
                  gid_url=event_url,
                  calibration = {},
                  interferometers=event_data['instruments'].split(","),
                  disable_repo=True
    )
    gitlab.EventIssue.create_issue(repository, event, issue_template="scripts/outline.md")


def find_calibrations(time):
    with open("../scripts/LLO_calibs.txt") as llo_file:
        data_llo = llo_file.read().split("\n")
        data_llo = [datum for datum in data_llo if datum[-16:]=="FinalResults.txt"]
        times_llo = {int(datum.split("GPSTime_")[1].split("_C01")[0]): datum for datum in data_llo}
    
    with open("../scripts/LHO_calibs.txt") as llo_file:
        data_lho = llo_file.read().split("\n")
        data_lho = [datum for datum in data_lho if datum[-16:]=="FinalResults.txt"]
        times_lho = {int(datum.split("GPSTime_")[1].split("_C01")[0]): datum for datum in data_lho}
        
    keys_llo = np.array(list(times_llo.keys())) 
    keys_lho = np.array(list(times_lho.keys())) 
    return {"LHO": times_lho[keys_lho[np.argmin(np.abs(keys_lho - time))]], "LLO": times_llo[keys_llo[np.argmin(np.abs(keys_llo - time))]]}

    
# Update existing events
for event in gitlab_events:
    data = client.superevent(event.title).json()
    event_data = client.event(data['preferred_event']).json()

    print(event_data['gpstime'])
    
    event.event_object.meta['event time'] = event_data['gpstime']
    event.update_data()



"""
These tests are designed to ensure that all classes of data in the
yaml 'blueprints' are correctly applied to the asimov ledger, and are
then correctly used to make config files.
"""

import os
import unittest
import shutil
import git
import asimov.event
from asimov.cli.project import make_project
from asimov.cli.application import apply_page
from asimov.ledger import YAMLLedger
from asimov.testing import AsimovTestCase


class DetcharTests(AsimovTestCase):
    """Tests to ensure that various detector characterisation related
    data are handled correctly.
    These should include:
    - minimum frequencies
    - maximum frequencies
    - data channels
    - frame types"""

    def test_event_non_standard_fmin(self):
        """Check event-specific fmin overwrites project default."""
        apply_page(
            f"{self.cwd}/tests/test_data/testing_pe.yaml",
            event=None,
            ledger=self.ledger,
        )
        apply_page(
            f"{self.cwd}/tests/test_data/event_non_standard_settings.yaml",
            event=None,
            ledger=self.ledger,
        )

        event = self.ledger.get_event("Nonstandard fmin")[0]

        self.assertEqual(event.meta["quality"]["minimum frequency"]["H1"], 62)
        self.assertEqual(event.meta["quality"]["minimum frequency"]["L1"], 92)
        self.assertEqual(event.meta["quality"]["minimum frequency"]["V1"], 62)

    def test_event_non_standard_channels(self):
        """Check event-specific channel overwrites project default."""
        apply_page(
            f"{self.cwd}/tests/test_data/testing_pe.yaml",
            event=None,
            ledger=self.ledger,
        )
        apply_page(
            f"{self.cwd}/tests/test_data/event_non_standard_settings.yaml",
            event=None,
            ledger=self.ledger,
        )

        event = self.ledger.get_event("Nonstandard fmin")[0]

        self.assertEqual(event.meta["data"]["channels"]["L1"], "L1:WeirdChannel")
        self.assertEqual(event.meta["data"]["channels"]["H1"], "H1:WeirdChannel")
        self.assertEqual(event.meta["data"]["channels"]["V1"], "V1:OddChannel")

    def test_event_non_standard_frames(self):
        """Check event-specific frame-type overwrites project default."""
        apply_page(
            f"{self.cwd}/tests/test_data/testing_pe.yaml",
            event=None,
            ledger=self.ledger,
        )
        apply_page(
            f"{self.cwd}/tests/test_data/event_non_standard_settings.yaml",
            event=None,
            ledger=self.ledger,
        )

        event = self.ledger.get_event("Nonstandard fmin")[0]

        self.assertEqual(event.meta["data"]["frame types"]["L1"], "NonstandardFrameL1")
        self.assertEqual(event.meta["data"]["frame types"]["H1"], "NonstandardFrame")
        self.assertEqual(event.meta["data"]["frame types"]["V1"], "UnusualFrameType")



"""
Provide the command line interface components for asimov.
"""

frametypes = {
    "L1": "L1_HOFT_CLEAN_SUB60HZ_C01",
    "H1": "H1_HOFT_CLEAN_SUB60HZ_C01",
    "V1": "V1Online",
}

CALIBRATION_NOTE = """
## Calibration envelopes
The following calibration envelopes have been found.
```yaml
---
{}
---
```
"""


ACTIVE_STATES = {"ready", "running", "stuck", "finished", "processing", "stop"}



import ast
import subprocess
from shutil import copy
from asimov import gitlab, mattermost
from asimov import config
from asimov import condor, git
from asimov.ini import RunConfiguration

import os, glob, datetime

import otter

import click

server = gitlab.gitlab.Gitlab('https://git.ligo.org', private_token=config.get("gitlab", "token"))
repository = server.projects.get(config.get("olivaw", "tracking_repository"))


uber_repository = git.MetaRepository(config.get("olivaw", "metarepository"))

events = gitlab.find_events(repository, milestone=config.get("olivaw", "milestone"))

mattermost = mattermost.Mattermost()
from pesummary.gw.file.read import read

def get_psds_rundir(rundir):
    psds = {}
    dets = ['L1', 'H1', 'V1']
    for det in dets:
        asset = f"{rundir}/ROQdata/0/BayesWave_PSD_{det}/post/clean/glitch_median_PSD_forLI_{det}.dat"
        if os.path.exists(asset):
            psds[det] = asset
    return psds


def upload_results(repo, event, prods, report):


    print("*"*15)
    print(f"Adding {prods} to the preferred summary file.")
    print("*"*15)

    samples = []
    labels = []
    configs = []
    
    calibration = []
    for prod in prods:
        rundir = event.data[f"{prod}_rundir"]
        samples.append(glob.glob(str(os.path.join(rundir, "posterior_samples"),)+"/*.hdf5")[0])

        run_ini = os.path.join(event.data[f"{prod}_rundir"], "config.ini")
        actual_config = RunConfiguration(run_ini)
        engine_data = actual_config.get_engine()
        psds = actual_config.get_psds()
        if len(psds) == 0:
            psds = get_psds_rundir(rundir)
        calib = actual_config.get_calibration()
        run_name = engine_data['approx'].replace('pseudoFourPN','')
        labels.append(f"C01:{run_name}")
        configs.append(str(os.path.join(event.data[f"{prod}_rundir"], "config.ini")))


    os.chdir(os.path.join(repo.directory, "Preferred", "PESummary_metafile"))

    command = ["summarycombine",
               "--webdir", f"/home/daniel.williams/public_html/LVC/projects/O3/preferred/{event.title}",
               "--samples", ]
    command += samples
    command += ["--labels"]
    command += labels
    command += ["--config"]
    command += configs
    command += ["--calibration"]
    command += [f"{key}:{value}" for key, value in calib.items()]
    command += ["--psd"]
    command += [f"{key}:{value}" for key, value in psds.items()]
    command += ["--gracedb"]
    command += [event.title]
    command += ["--redshift_method",  "exact"]
    command += ["--cosmology", "Planck15_lal"]
    command += ["--gw"]

    print(command)
    with report:
        report + f"```\n {command}```"

    dagman = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    out, err = dagman.communicate()

    print(out)
    print(err)

    copy(f"/home/daniel.williams/public_html/LVC/projects/O3/preferred/{event.title}/samples/posterior_samples.h5", os.path.join(repo.directory, "Preferred", "PESummary_metafile"))
    repo.repo.git.add("Preferred/PESummary_metafile/posterior_samples.h5")
    repo.repo.git.commit("-m", "Updated the preferred sample metafile.")
    repo.repo.git.push()

    event.labels += ["Preferred cleaned"]
    event.issue_object.save()

    return True

import click

@click.command()
@click.option('--event', "eventflag", default=None, help='The event to rebuild')
def main(eventflag):

    report = otter.Otter(filename="/home/daniel.williams/public_html/LVC/projects/O3/C01/audit_current.html", 
                         author="R. Daniel Williams",
                         title="Asimov/Olivaw : Preferred run audit report"
    )

    with report:
        report + "This report contains the latest updates on the run status of the various PE runs ongoing at CIT."
        report + "Supervisor last checked these events at " + str(datetime.datetime.now())

    global events


    if eventflag:
        events = [event for event in events if event.title == eventflag]

    for event in events:
        print(event.title)

        # if event.state == None:
        #     continue
        # if "Special" in event.labels:
        #    continue
        # if "Preferred cleaned" in event.labels:
        #    continue

        try:
            repo = uber_repository.get_repo(event.title)
        except:
            print(f"{event.title} missing from the uberrepo")
            continue

        repo.update(stash=True)

        with report:
            report + f"#{event.title}"

        preferred_summary = os.path.join(repo.directory, "Preferred", "PESummary_metafile", "posterior_samples.h5")

        try:
            print(preferred_summary)
            data = read(preferred_summary)

            print(data.labels)

            with report:
                report + f"{data.labels}"
        except OSError:
            with report:
                report + f"There is no preferred file in this repository."


        try:
            event_prods = repo.find_prods(config.get("olivaw", "run_category"))
        except:
            print(f"No runs in this repository")
            continue

        if event.state in ["Generating PSDs", "Productions running", "Stuck"]:
            psds_dict = {}
            prod_keys = [key for key in event.data.keys() if "Prod" in key[0:5]]
            n_productions = len(event_prods)
            event_psds_l = []

            pref_prod = []
            for prod in event_prods:
                prod = prod.split("_")[0]
                if prod in event.data:
                    cluster = event.data[prod]

                    run_ini = os.path.join(event.data[f"{prod}_rundir"], "config.ini")
                    actual_config = RunConfiguration(run_ini)
                    engine_data = actual_config.get_engine()

                    if "finalised" in cluster.lower():
                        print(f"{prod} is the preferred production")
                        pref_prod.append(prod)
            try:
                if len(pref_prod)==0: continue
                upload_results(repo, event, pref_prod, report=report)
            except FileNotFoundError as e:
                print(e)

if __name__ == "__main__":

   main()



"""
Code to handle the various kinds of analysis which asimov can handle.

Asimov defintes three types of analysis, depending on the inputs of the analysis.

Simple analyses
  These analyses operate only on a single event,
  and will generally use a very specific set of configuration settings.
  An example of a simple analysis is a bilby or RIFT parameter estimation analysis,
  as these only require access to the data for a single event.
  Before version 0.4 these were called `Productions`.

Event analyses
  These analyses can access the results of all of the simple analyses which have been
  performed on a single event, or a subset of them.
  An example of an event analysis is the production of mixed posterior samples from multiple
  PE analyses.

Project analyses
  These are the most general type of analysis, and have access to the results of all analyses
  on all events, including event and simple analyses.
  This type of analysis is useful for defining a population analysis, for example.

"""
import os
import configparser
from copy import deepcopy
from warnings import warn

from liquid import Liquid

from asimov import config
from asimov.pipelines import known_pipelines
from asimov.utils import update
from asimov.storage import Store


class Analysis:
    """
    The base class for all other types of analysis.

    TODO: Add a check to make sure names cannot conflict
    """

    @property
    def review(self):
        """
        Return the review information attached to the analysis.
        """

    def _process_dependencies(self, needs):
        """
        Process the dependencies list for this production.
        """
        return needs

    @property
    def dependencies(self):
        if "needs" in self.meta:
            dependencies = self._process_dependencies(self.meta["needs"])
        else:
            dependencies = None
        return dependencies

    @property
    def priors(self):
        if "priors" in self.meta:
            priors = self.meta["priors"]
        else:
            priors = None
        return priors

    @property
    def finished(self):
        finished_states = ["uploaded"]
        return self.status in finished_states

    @property
    def status(self):
        return self.status_str.lower()

    @status.setter
    def status(self, value):
        self.status_str = value.lower()
        if self.event.issue_object is not None:
            self.event.issue_object.update_data()

    def results(self, filename=None, handle=False, hash=None):
        store = Store(root=config.get("storage", "results_store"))
        if not filename:
            try:
                items = store.manifest.list_resources(self.subject.name, self.name)
                return items
            except KeyError:
                return None
        elif handle:
            return open(
                store.fetch_file(self.subject.name, self.name, filename, hash), "r"
            )
        else:
            return store.fetch_file(self.subject.name, self.name, filename, hash=hash)

    @property
    def rundir(self):
        """
        Return the run directory for this event.
        """
        if "rundir" in self.meta:
            return self.meta["rundir"]
        elif "working directory" in self.subject.meta:
            value = os.path.join(self.subject.meta["working directory"], self.name)
            self.meta["rundir"] = value
            # TODO: Make sure this is saved back to the ledger
        else:
            return None

    def make_config(self, filename, template_directory=None):
        """
        Make the configuration file for this production.

        Parameters
        ----------
        filename : str
           The location at which the config file should be saved.
        template_directory : str, optional
           The path to the directory containing the pipeline config templates.
           Defaults to the directory specified in the asimov configuration file.
        """

        if "template" in self.meta:
            template = f"{self.meta['template']}.ini"
        else:
            template = f"{self.pipeline}.ini"

        pipeline = known_pipelines[self.pipeline]

        try:
            template_directory = config.get("templating", "directory")
            template_file = os.path.join(f"{template_directory}", template)

        except (configparser.NoOptionError, configparser.NoSectionError):
            if hasattr(pipeline, "config_template"):
                template_file = pipeline.config_template
            else:
                from pkg_resources import resource_filename

                template_file = resource_filename("asimov", f"configs/{template}")

        liq = Liquid(template_file)
        rendered = liq.render(production=self, config=config)

        with open(filename, "w") as output_file:
            output_file.write(rendered)


class SimpleAnalysis(Analysis):
    """
    A single subject, single pipeline analysis.
    """

    def __init__(self, subject, name, pipeline, status=None, comment=None, **kwargs):
        self.event = self.subject = subject
        self.name = name

        if status:
            self.status_str = status.lower()
        else:
            self.status_str = "none"
            self.pipeline = pipeline.lower()
            self.comment = comment
            self.meta = deepcopy(self.subject.meta)
            self.meta = update(self.meta, kwargs)


class SubjectAnalysis(Analysis):
    """
    A single subject analysis which requires results from multiple pipelines.
    """

    pass


class ProjectAnalysis(Analysis):
    """
    A multi-subject analysis.
    """

    def __init__(
        self, subjects, analyses, name, pipeline, status=None, comment=None, **kwargs
    ):
        """ """
        super().__init__()

        self.subjects = subjects
        self.analyses = analyses


class GravitationalWaveTransient(SimpleAnalysis):
    """
    A single subject, single pipeline analysis for a gravitational wave transient.
    """

    def __init__(self, subject, name, pipeline, status=None, comment=None, **kwargs):
        super().init(subject, name, pipeline, status=None, comment=None, **kwargs)
        self._checks()

    def _checks(self):
        """
        Carry-out a number of data consistency checks on the information from the ledger.
        """
        # Check that the upper frequency is included, otherwise calculate it
        if self.quality:
            if ("high-frequency" not in self.quality) and (
                "sample-rate" in self.quality
            ):
                # Account for the PSD roll-off with the 0.875 factor
                self.meta["quality"]["high-frequency"] = int(
                    0.875 * self.meta["quality"]["sample-rate"] / 2
                )
            elif ("high-frequency" in self.quality) and ("sample-rate" in self.quality):
                if self.meta["quality"]["high-frequency"] != int(
                    0.875 * self.meta["quality"]["sample-rate"] / 2
                ):
                    warn(
                        "The upper-cutoff frequency is not equal to 0.875 times the Nyquist frequency."
                    )

    @property
    def quality(self):
        if "quality" in self.meta:
            return self.meta["quality"]
        else:
            return None

    @property
    def reference_frame(self):
        """
        Calculate the appropriate reference frame.
        """
        ifos = self.meta["interferometers"]
        if len(ifos) == 1:
            return ifos[0]
        else:
            return "".join(ifos[:2])

    @property
    def psds(self):
        """
        Return the PSDs stored for this transient event.
        """
        if "psds" in self.meta and self.quality:
            if self.quality["sample-rate"] in self.meta["psds"]:
                self.psds = self.meta["psds"][self.quality["sample-rate"]]
            else:
                self.psds = {}
        else:
            self.psds = {}

        for ifo, psd in self.psds.items():
            if self.subject.repository:
                self.psds[ifo] = os.path.join(self.subject.repository.directory, psd)
            else:
                self.psds[ifo] = psd

    def get_timefile(self):
        """
        Find this event's time file.

        Returns
        -------
        str
           The location of the time file.
        """
        return self.event.repository.find_timefile(self.category)

    def get_coincfile(self):
        """
        Find this event's coinc.xml file.

        Returns
        -------
        str
           The location of the time file.
        """
        try:
            coinc = self.event.repository.find_coincfile(self.category)
            return coinc
        except FileNotFoundError:
            self.event.get_gracedb(
                "coinc.xml",
                os.path.join(
                    self.event.repository.directory, self.category, "coinc.xml"
                ),
            )
            coinc = self.event.repository.find_coincfile(self.category)
            return coinc


class Production(SimpleAnalysis):
    pass



"""
The asimov storage API.

Asimov handles the results files for productions using a nested directory layout.
This API helps to make using this easier.


Configuration options
---------------------

Configuration options for the storage API can be found in the `storage` namespace.

```
[storage]
root = /path/to/the/storage/root

```

"""

import hashlib
import os
import pathlib
import stat
import uuid
from shutil import copyfile

import yaml


class NotAStoreError(Exception):
    pass


class AlreadyPresentException(Exception):
    pass


class HashError(Exception):
    pass


class Manifest:
    """
    The storage manifest.

    This object contains details of all of the resources (files) stored
    in a Store, including auditing information such as hashes and UUIDs.

    Parameters
    ----------
    store : `asimov.storage.Store`
       The results store which this should be the manifest of.
    """

    uuid_hash = None

    def __init__(self, store):
        self.root = store.root
        self.store = store

        try:
            with self._open() as f:
                self.data = yaml.safe_load(f)
        except FileNotFoundError:
            raise NotAStoreError

    @property
    def hash_dict(self):
        data = {}
        for event in self.data["events"].values():
            for production in event.values():
                for resource in production.values():
                    data[resource["hash"]] = resource
        return data

    @property
    def uuid_dict(self):
        data = {}
        for e_name, event in self.data["events"].items():
            for p_name, production in event.items():
                for r_name, resource in production.items():
                    data[resource["uuid"]] = os.path.join(
                        self.root, e_name, p_name, resource["uuid"]
                    )
        return data

    def get_hash(self, uuid):
        if not self.uuid_hash:
            self.uuid_hash = {}
            for e_name, event in self.data["events"].items():
                for p_name, production in event.items():
                    for r_name, resource in production.items():
                        self.uuid_hash[resource["uuid"]] = resource["hash"]
        return self.uuid_hash[uuid]

    def _open(self):
        """
        Open the manifest file.
        """
        manifest = os.path.join(self.root, ".manifest", "manifest.yaml")
        # Check the manifest exists
        if os.path.isfile(manifest):
            return open(manifest, "r")
        else:
            raise FileNotFoundError

    def update(self):
        """
        Update the manifest file.
        """
        manifest = os.path.join(self.root, ".manifest", "manifest.yaml")
        if os.path.isfile(manifest):
            with open(manifest, "w") as f:
                f.write(yaml.dump(self.data))
        else:
            raise FileNotFoundError

    def get_uuid(self, event, production, filename):
        try:
            return self.data["events"][event][production][filename]["uuid"]
        except KeyError:
            raise FileNotFoundError

    @classmethod
    def create(cls, store):
        """
        Create the manifest.
        This should only be run on a new store, and will fail if a `.manifest` directory exists
        in the Store already.
        """
        contents = {"name": store["name"], "events": {}}
        manifest = os.path.join(store["root"], ".manifest", "manifest.yaml")

        if not os.path.isfile(manifest):
            with open(manifest, "w") as f:
                f.write(yaml.dump(contents))
        else:
            raise FileExistsError

    def _verify(self, hash):
        """
        Verify the manifest directory's hash.

        Parameters
        ----------
        hash : str
           The hash of the manifest directory itself.
        """
        pass

    def list_resources(self, event, production):
        """
        List all of the resources available for a production.
        """
        return self.data["events"][event][production]

    def add_record(self, event, production, resource, hash, resource_uuid):
        """
        Add a resource record to the manifest.

        Parameters
        ----------
        event : `asimov.events.Event` or str
           The event which this file is for.
           Can be either an event object, or the name of the event.
        production : `asimov.events.Production` or str
           Can be either a production object or the name of the production.
        resource : str
           The path to the resource to be recorded.
        hash : str
           The hash of the resource being recorded.
        uuid : UUID
           The UUID of the object being recorded.
        """
        # This function should store the name, location, event, production of the file
        # then calculate the hash and uuid for the file, and store them in the manifest

        if event not in self.data["events"]:
            self.data["events"][event] = {}
        if production not in self.data["events"][event]:
            self.data["events"][event][production] = {}

        if resource in self.data["events"][event][production]:
            raise FileExistsError

        self.data["events"][event][production][resource] = {
            "uuid": resource_uuid.hex,
            "hash": hash,
        }


class Store:
    """
    The results store.

    Parameters
    ----------
    root : str, optional
       The path to the results store.
       If not provided defaults to the value of `storage>root` in the configuration file.
    """

    def __init__(self, root=None):
        """
        Initiate an asimov store.
        """
        self.root = root

        self.manifest = Manifest(self)

    @classmethod
    def create(cls, root, name):
        """
        Create this results store.

        Parameters
        ----------
        root : str
           The path to the results store.
        name : str
           A name for this Store.
        """
        pathlib.Path(root).mkdir(parents=False, exist_ok=False)
        manifest_dir = os.path.join(root, ".manifest")
        pathlib.Path(manifest_dir).mkdir(parents=False, exist_ok=False)
        store = {}
        store["name"] = name
        store["root"] = root
        _ = Manifest.create(store)

    def _check(self):
        """
        Check the integrity of the results store.
        """
        pass

    def _exists(self):
        """
        Check if the store exists.
        """
        pass

    def _hash(self, path):
        """
        Calculate the hash of a file.

        Parameters
        ----------
        path : str
           The filepath of the file to be hashed.
        """
        hasher = hashlib.md5()
        with open(path, "rb") as afile:
            buf = afile.read()
            hasher.update(buf)
        return hasher.hexdigest()

    def add_file(self, event, production, file, new_name=None):
        """
        Add a file to the store.

        Parameters
        ----------
        event : `asimov.events.Event` or str
           The event which this file is for.
           Can be either an event object, or the name of the event.
        production : `asimov.events.Production` or str
           Can be either a production object or the name of the production.
        file : str
           The path to the origin file to be stored.
        new_name : str
           The new name of the file once it's stored.

        Returns
        -------
        hash : str
           The MD5 hash of the stored file.
        """
        hash = self._hash(file)
        if hash in self.manifest.hash_dict:
            raise AlreadyPresentException

        pathlib.Path(os.path.join(self.root, event, production)).mkdir(
            parents=True, exist_ok=True
        )

        this_uuid = uuid.uuid4()

        name = new_name if new_name else os.path.basename(file)

        self.manifest.add_record(event, production, name, hash, this_uuid)

        destination = os.path.join(self.root, event, production, name)

        copyfile(file, destination)

        os.chmod(destination, stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH)

        self.manifest.update()

        return {"file": name, "hash": self._hash(destination), "uuid": this_uuid.urn}

    def fetch_file(self, event, production, file, hash=None):
        """
        Retrieve a file from the store.

        Parameters
        ----------
        event : `asimov.events.Event` or str
           The event which this file is for.
           Can be either an event object, or the name of the event.
        production : `asimov.events.Production` or str
           Can be either a production object or the name of the production.
        file : str
           The name of the file to be retrieved.
        hash : str, optional
           The expected MD5 hash of the file.
           If this is not provided the file will be verified against the store's manifest
           before being returned, but if a hash is provided it will be checked against
           the provided value.

        Returns
        -------
        path : str
           The path to the file.
        """
        this_uuid = self.manifest.get_uuid(event, production, file)
        resource = self.fetch_uuid(this_uuid)
        stored_hash = self.manifest.get_hash(this_uuid)
        file_hash = self._hash(resource)

        if file_hash != stored_hash:
            raise HashError(
                "The file in the file store's hash does not match the manifest."
            )

        if hash:
            if hash != stored_hash:
                raise HashError("The manifest hash does not match the check hash.")

        return resource

    def fetch_uuid(self, uuid):
        """
        Retrieve a file from the store from its uuid.

        Parameters
        ----------
        uuid : str
           The uuid of the requested resource.

        Returns
        -------
        file : str
           The path to the file.
        """
        return self.manifest.uuid_dict[uuid]



"""
Test suites for the implementation of analyses.

Analyses replace and generalise the earlier notion of a "production" from v0.3 and earlier.
"""

import unittest

class TestBaseAnalysis(unittest.TestCase):
    pass

class TestSimpleAnalysis(TestBaseAnalysis):
    pass

class TestEventAnalysis(TestBaseAnalysis):
    pass

class TestProjectAnalysis(TestBaseAnalysis):
    pass



from setuptools import setup

with open('README.md') as readme_file:
    readme = readme_file.read()

# with open('HISTORY.rst') as history_file:
#     history = history_file.read()

with open("requirements.txt") as requires_file:
    requirements = requires_file.read().split("\n")

requirements = [requirement for requirement in requirements if not ("+" in requirement)]
    
test_requirements = [
    # TODO: put package test requirements here
]

setup(
    name='asimov',
    use_scm_version={"local_scheme": "no-local-version"},
    setup_requires=['setuptools_scm'],
    description="""A Python package for managing and interacting with data analysis jobs.""",
    long_description=readme,
    long_description_content_type="text/markdown",
    author="Daniel Williams",
    author_email='daniel.williams@ligo.org',
    url='https://git.ligo.org/asimov/asimov',
    packages=['asimov'],
    package_dir={'asimov': 'asimov'},
    entry_points={
        'console_scripts': [
            'olivaw=asimov.olivaw:olivaw',
            'asimov=asimov.olivaw:olivaw',
            'locutus=asimov.locutus:cli'
        ]
    },
    include_package_data=True,
    install_requires=requirements,
    license="MIT license",
    zip_safe=True,
    keywords='pe, ligo, asimov',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Astronomy',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'License :: OSI Approved :: ISC License (ISCL)',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
    test_suite='tests',
    tests_require=test_requirements,
)



"""
Code for interacting with a gitlab instance.
"""
import datetime
import time
import configparser

import gitlab
import yaml
from liquid import Liquid

from asimov import config

from .event import Event
from .ledger import Ledger

STATE_PREFIX = "C01"


class GitlabLedger(Ledger):
    """
    Connect to a gitlab-based issue tracker.
    """

    def __init__(
        self, configs, milestone=None, subset=None, update=False, repo=True, label=None
    ):
        """
        Search through a repository's issues and find all of the ones
        for events.
        """

        _, self.repository = self._connect_gitlab()

        if subset == [None]:
            subset = None
        if not label:
            event_label = config.get("gitlab", "event_label")
        else:
            event_label = label
        try:
            sleep_time = int(config.get("gitlab", "rest_time"))
        except configparser.NoOptionError:
            sleep_time = 30
        issues = self.repository.issues.list(labels=[event_label], per_page=1000)
        output = []
        if subset:
            for issue in issues:
                if issue.title in subset:
                    output += [EventIssue(issue, self.repository, update, repo=repo)]
                    if update:
                        time.sleep(sleep_time)
        else:
            for issue in issues:
                output += [EventIssue(issue, self.repository, update, repo=repo)]
                if update:
                    time.sleep(sleep_time)

        self.data["events"] = output
        self.events = {ev["name"]: ev for ev in self.data["events"]}

    def _connect_gitlab(self):
        """
        Connect to the gitlab server.

        Returns
        -------
        server : `Gitlab`
           The gitlab server.
        repository: `Gitlab.project`
           The gitlab project.
        """
        server = gitlab.gitlab.Gitlab(
            config.get("gitlab", "server"), private_token=config.get("gitlab", "token")
        )
        repository = server.projects.get(config.get("gitlab", "tracking_repository"))
        return server, repository

    def get_event(self, event=None):
        if event:
            return self.events[event]
        else:
            return self.events.values()

    @classmethod
    def add_event(self, event_object, issue_template=None):
        """
        Create an issue for an event.
        """

        if not issue_template:
            from pkg_resources import resource_filename

            issue_template = resource_filename("asimov", "gitlabissue.md")

        with open(issue_template, "r") as template_file:
            liq = Liquid(template_file.read())
            rendered = liq.render(
                event_object=event_object, yaml=event_object.to_yaml()
            )

        self.repository.issues.create(
            {"title": event_object.name, "description": rendered}
        )

    def update_event(self, event):
        event.update_data()

    def save(self):
        pass

    @classmethod
    def create(cls):
        raise NotImplementedError(
            "You must create a gitlab issue tracker manually first."
        )

    pass


class EventIssue(Event):
    """
    Use an issue on the gitlab issue tracker to
    hold variable data for the program.

    Parameters
    ----------
    issue : `gitlab.issue`
       The issue which represents the event.
    update : bool
       Flag to determine if the git repository is updated
       when it is loaded. Defaults to False to prevent
       excessive load on the git server.
    """

    def __init__(self, issue, repository, update=False, repo=True):

        self.from_issue(self, issue, update, repo)

        self.issue_object = issue
        self.title = issue.title
        self.text = issue.description

        self.issue_id = issue.id
        self.labels = issue.labels
        self.data = self.parse_notes()

    def from_issue(self, issue, update=False, repo=True):
        """
        Parse an issue description to generate this event.


        Parameters
        ----------
        update : bool
           Flag to determine if the repository is updated when loaded.
           Defaults to False.
        """

        text = issue.text.split("---")

        event = self.from_yaml(text[1], issue, update=update, repo=repo)
        event.text = text
        # event.from_notes()

        return event

    def _refresh(self):
        if self.repository:
            self.issue_object = self.repository.issues.get(self.issue_object.iid)
            self.text = self.issue_object.description.split("---")
        else:
            pass

    @property
    def state(self):
        """
        Get the state of the event's runs.
        """
        for label in self.issue_object.labels:
            if f"{STATE_PREFIX}::" in label:
                return label[len(STATE_PREFIX) + 2:]
        return None

    @state.setter
    def state(self, state):
        """
        Set the event state.
        """
        self._refresh()
        for label in self.issue_object.labels:
            if f"{STATE_PREFIX}::" in label:
                # Need to remove all of the other scoped labels first.
                self.issue_object.labels.remove(label)
        self.issue_object.labels += ["{}::{}".format(STATE_PREFIX, state)]
        self.issue_object.save()

    def add_note(self, text):
        """
        Add a comment to the event issue.
        A footer will be added to identify this as being created by the
        supervisor and not the user.
        """
        self._refresh()
        now = datetime.datetime.now()
        header = """"""
        footer = f"""\nAdded at {now:%H}:{now:%M}, {now:%Y}-{now:%m}-{now:%d} by the run supervision robot :robot:."""
        self.issue_object.notes.create({"body": header + text + footer})
        self.issue_object.save()

    def add_label(self, label, state=True):
        """
        Add a new label to an event issue.

        Parameters
        ----------
        label : str
           The name of the label.
        """
        self._refresh()
        if state:
            self.issue_object.labels += [f"{STATE_PREFIX}:{label}"]
        else:
            self.issue_object.labels += [f"{label}"]

        self.issue_object.save()

    def update_data(self):
        """
        Store event data in the comments on the event repository.
        """
        self._refresh()

        self.issue_object.description = self.event_object.to_issue()

        self.issue_object.save()

    def parse_notes(self):
        """
        Read issue information from the comments on the issue.

        Only notes which start
        ```
        # Run information
        ```
        will be parsed.
        """
        data = {}
        notes = self.issue_object.notes.list(per_page=200)
        note_data = []
        for note in reversed(notes):
            if "---\n" in note.body:
                data = note.body.split("---")
                if len(data) > 0:
                    data = data[1]
                else:
                    continue
                data = yaml.safe_load(data)
                note_data.append(data)
        return note_data
